from misc import get_args
import pytest

import sys
import os
here = os.path.dirname(__file__)

sys.path.append(os.path.join(here, '..'))

@pytest.mark.parametrize("embedding, backbone", [("feature_tokenizer", "transformer")])
def test_switchtab_classification(embedding, backbone):
    
    from benchmark.datasets import load_diabetes
    data, label, continuous_cols, category_cols, output_dim, metric, metric_hparams = load_diabetes()
    
    sys.path.append(os.path.join(os.path.join(here, '..'), "benchmark"))
    
    from benchmark.pipelines import SwitchTabPipeLine
    
    args = get_args()
    args.embedding = embedding
    args.backbone = backbone
    
    pipeline = SwitchTabPipeLine(args, data, label, continuous_cols, category_cols, output_dim, metric, metric_hparams)
    
    pipeline.benchmark()

@pytest.mark.parametrize("embedding, backbone", [("feature_tokenizer", "transformer")])
def test_switchtab_regression(embedding, backbone):
    
    from benchmark.datasets import load_abalone
    data, label, continuous_cols, category_cols, output_dim, metric, metric_hparams = load_abalone()
    
    sys.path.append(os.path.join(os.path.join(here, '..'), "benchmark"))
    
    from benchmark.pipelines import SwitchTabPipeLine
    
    args = get_args()
    args.embedding = embedding
    args.backbone = backbone
    
    pipeline = SwitchTabPipeLine(args, data, label, continuous_cols, category_cols, output_dim, metric, metric_hparams)
    
    pipeline.benchmark()
    
if __name__ == "__main__":
    test_switchtab_classification("feature_tokenizer", "transformer")
    test_switchtab_regression("feature_tokenizer", "transformer")


from typing import OrderedDict, Tuple

import torch
import torch.nn as nn
import torch.nn.functional as F

from ts3l.models.common import TS3LModule
from ts3l.models.common import MLP
from ts3l.utils import BaseEmbeddingConfig, BaseBackboneConfig

class SCARF(TS3LModule):
    def __init__(
        self,
        embedding_config: BaseEmbeddingConfig,
        backbone_config: BaseBackboneConfig,
        pretraining_head_dim: int,
        output_dim: int,
        head_depth: int =2,
        dropout_rate: float = 0.04,
        **kwargs
    ) -> None:
        """Implementation of SCARF: Self-Supervised Contrastive Learning using Random Feature Corruption.
        It consists in an encoder that learns the embeddings.
        It is done by minimizing the contrastive loss of a sample and a corrupted view of it.
        The corrupted view is built by replacing a random set of features by another sample randomly drawn independently.
            Args:
                embedding_config (BaseEmbeddingConfig): Configuration for the embedding layer.
                backbone_config (BaseBackboneConfig): Configuration for the backbone network.
                pretraining_head_dim (int): The dimensionality of the pretraining head.
                output_dim (int): The dimensionality of output.
                head_depth (int, optional): The number of layers of the pretraining head. Defaults to 2.
                dropout_rate (float, optional): A hyperparameter that is to control dropout layer of the head. Default is 0.04.
        """
        super(SCARF, self).__init__(embedding_config, backbone_config)

        self.pretraining_head = MLP(input_dim = self.backbone_module.output_dim, hidden_dims=pretraining_head_dim, n_hiddens=head_depth, dropout_rate=dropout_rate)

        self.head = nn.Sequential(
            OrderedDict([
                ("head_activation", nn.ReLU(inplace=True)),
                ("head_batchnorm", nn.BatchNorm1d(self.backbone_module.output_dim)),
                ("head_dropout", nn.Dropout(dropout_rate)),
                ("head_linear", nn.Linear(self.backbone_module.output_dim, output_dim))
            ])
        )

    @property
    def encoder(self) -> nn.Module:
        return self.backbone_module
        
    def _first_phase_step(self, x: torch.Tensor, x_corrupted: torch.Tensor) -> Tuple[torch.Tensor, torch.Tensor]:
        
        size = len(x)
        x = torch.concat((x, x_corrupted))
        
        x = self.embedding_module(x)
        emb_anchor = self.encoder(x)
        
        emb_anchor = self.pretraining_head(emb_anchor)
        emb_anchor = F.normalize(emb_anchor, p=2)
        
        return emb_anchor[:size], emb_anchor[size:]
    
    def _second_phase_step(self, x) -> torch.Tensor:
        x = self.embedding_module(x)
        emb = self.encoder(x)
        output = self.head(emb)

        return output


from dataclasses import dataclass, field
from ts3l.utils import BaseConfig

from typing import Any, List

@dataclass
class SCARFConfig(BaseConfig):
    """ Configuration class for initializing components of the SCARFLightning Module, including hyperparameters of SCARF,
    optimizers, learning rate schedulers, and loss functions, along with their respective hyperparameters.

    Inherits Attributes:
        task (str): Specify whether the problem is regression or classification.
        embedding_config (BaseEmbeddingConfig): Configuration for the embedding layer.
        backbone_config (BaseBackboneConfig): Configuration for the backbone network.
        output_dim (int): The dimension of output.
        loss_fn (str): Name of the loss function to be used. Must be an attribute of 'torch.nn'.
        loss_hparams (Dict[str, Any]): Hyperparameters for the loss function. Default is empty dictionary.
        metric (str): Name of the metric to be used. Must be an attribute of 'torchmetrics.functional' or 'sklearn.metrics'. Default is None.
        metric_hparams (Dict[str, Any]): Hyperparameters for the metric. Default is an empty dictionary.
        optim (str): Name of the optimizer to be used. Must be an attribute of 'torch.optim'. Default is 'AdamW'.
        optim_hparams (Dict[str, Any]): Hyperparameters for the optimizer. Default is {'lr': 0.0001, 'weight_decay': 0.00005}.
        scheduler (str): Name of the learning rate scheduler to be used. Must be an attribute of 'torch.optim.lr_scheduler' or None. Default is None.
        scheduler_hparams (Dict[str, Any]): Hyperparameters for the scheduler. Default is None, indicating no scheduler is used.
        initialization (str): The way to initialize neural network parameters. Default is 'kaiming_uniform'.
        random_seed (int): Seed for random number generators to ensure reproducibility. Defaults to 42.
        
    New Attributes:
        pretraining_head_dim (int): The dimension of the pretraining head. Default is 256.
        head_depth (bool): The depth of head. Default is 2.
        dropout_rate (bool): A hyperparameter that is to control dropout layer. Default is 0.04.
        tau (float): A hyperparameter that is to scale similarity between views during the first phase.
        corruption_rate (float): The proportion of features to be corrupted, simulating noisy conditions for robustness. 
                For the second phase dataset, it should be 0. Defaults to 0.0.

    Raises:
        ValueError: Inherited from 'BaseConfig' to indicate that a configuration for the task, optimizer, scheduler, loss function, or metric is either invalid or not specified.
        
    """
    
    pretraining_head_dim: int = field(default=256)
    
    head_depth: int = field(default=2)
    
    dropout_rate: float = field(default=0.04)
    
    tau: float = field(default=0.1)
    
    corruption_rate: float = field(default=0)
    
    def __post_init__(self):
        super().__post_init__()


from typing import Dict, Any, List, Type
from ts3l.utils import RegressionMetric, ClassificationMetric

from abc import ABC, abstractmethod

import torch
from torch import nn

import pytorch_lightning as pl

from dataclasses import asdict
from ts3l.utils import BaseConfig
from ts3l.models.common import initialize_weights

    
class TS3LLightining(ABC, pl.LightningModule):
    """The pytorch lightning module of TabularS3L
    """
    def __init__(self, config: BaseConfig) -> None:
        """Initialize the pytorch lightining module of TabularS3L

        Args:
            config (BaseConfig): The configuration of TS3LLightining.
        """
        super(TS3LLightining, self).__init__()
        
        self.random_seed = config.random_seed
        
        pl.seed_everything(self.random_seed)
        
        self.optim = getattr(torch.optim, config.optim)
        self.optim_hparams = config.optim_hparams
        
        self.sched = getattr(torch.optim.lr_scheduler, config.scheduler) if config.scheduler is not None else None
        self.scheduler_hparams = config.scheduler_hparams
        
        self.task_loss_fn = getattr(torch.nn, config.loss_fn)(**config.loss_hparams)
        
        self.__configure_metric(config.task, config.metric, config.metric_hparams)
        
        self._initialize(config)
        
        self.set_first_phase()

        self.first_phase_step_outputs: List[Dict[str, Any]] = []
        self.second_phase_step_outputs: List[Dict[str, Any]] = []
        
        self.save_hyperparameters()
    
    @abstractmethod
    def _initialize(self, config: BaseConfig) -> None:
        pass
    
    def _init_model(self, model_class: Type[nn.Module], config: BaseConfig) -> None:
        initialization = config.initialization
        
        self.model = model_class(**config.__dict__)
        initialize_weights(self.model, initialization)
        
    def __configure_metric(self, task, metric, metric_hparams):
        
        if task == "regression":
            self.metric = RegressionMetric(metric, metric_hparams)
        else:
            self.metric = ClassificationMetric(metric, metric_hparams)
            
    def configure_optimizers(self):
        """Configure the optimizer
        """
        self.optimizer = self.optim(self.parameters(), **self.optim_hparams)
        if self.sched is None:
            return [self.optimizer]
        self.scheduler = self.sched(self.optimizer, **self.scheduler_hparams)
        return [self.optimizer], [{'scheduler': self.scheduler, 'interval': 'step'} ]

    def set_first_phase(self) -> None:
        """Set the module to pretraining
        """
        self.model.set_first_phase()
        self.training_step = self._first_phase_step # type: ignore
        self.on_validation_start = self._on_first_phase_validation_start # type: ignore
        self.validation_step = self._first_phase_step # type: ignore
        self.on_validation_epoch_end = self._first_phase_validation_epoch_end # type: ignore

    def set_second_phase(self, freeze_encoder: bool) -> None:
        """Set the module to fine-tuning
        
        Args:
            freeze_encoder (bool): If True, the encoder will be frozen during fine-tuning. Otherwise, the encoder will be trainable.
        """
        self.model.set_second_phase(freeze_encoder)
        self.training_step = self._second_phase_step # type: ignore
        self.on_validation_start = self._on_second_phase_validation_start # type: ignore
        self.validation_step = self._second_phase_step # type: ignore
        self.on_validation_epoch_end = self._second_phase_validation_epoch_end # type: ignore

    def forward(self,
                batch:Dict[str, Any]
    ) -> torch.Tensor:
        """Do forward pass for given input

        Args:
            batch (Dict[str, Any]): The input batch

        Returns:
            torch.Tensor: The output of forward pass
        """
        return self.model(batch)
    

    @abstractmethod
    def _get_first_phase_loss(self, batch: Any) -> torch.Tensor:
        """Calculate the first phase loss

        Args:
            batch (Any): The input batch

        Returns:
            torch.Tensor: The final loss of first phase step
        """
        pass
    
    def _first_phase_step(self,
                      batch: Any,
                      batch_idx: int
    ) -> Dict[str, Any]:
        """The first phase step of TabularS3L

        Args:
            batch (Any): The input batch
            batch_idx (int): Only for compatibility

        Returns:
            Dict[str, Any]: The loss of the first phase step
        """

        loss = self._get_first_phase_loss(batch)
        self.first_phase_step_outputs.append({
            "loss" : loss
        })
        return {
            "loss" : loss
        }

    def _on_first_phase_validation_start(self):
        """Log the training loss of the first_phase
        """
        if len(self.first_phase_step_outputs) > 0:
            train_loss = torch.Tensor([out["loss"] for out in self.first_phase_step_outputs]).cpu().mean()
            
            self.log("train_loss", train_loss, prog_bar = True)
            
            self.first_phase_step_outputs = []    
        return super().on_validation_start() 
    
    def _first_phase_validation_epoch_end(self) -> None:
        """Log the validation loss of the first phase
        """
        val_loss = torch.Tensor([out["loss"] for out in self.first_phase_step_outputs]).cpu().mean()

        self.log("val_loss", val_loss, prog_bar = True)
        self.first_phase_step_outputs = []
        return super().on_validation_epoch_end()

    @abstractmethod
    def _get_second_phase_loss(self, batch: Any):
        """Calculate the second phase loss

        Args:
            batch (Any): The input batch

        Returns:
            torch.FloatTensor: The final loss of second phase step
            torch.Tensor: The label of the labeled data
            torch.Tensor: The predicted label of the labeled data
        """
        pass
        
    
    def _second_phase_step(self,
                      batch: Any,
                      batch_idx: int = 0
    ) -> Dict[str, Any]:
        """The second phase step of TabularS3L

        Args:
            batch (Any): The input batch
            batch_idx (int): Only for compatibility

        Returns:
            Dict[str, Any]: The loss of the second phase step
        """
        loss, y, y_hat = self._get_second_phase_loss(batch)

        self.second_phase_step_outputs.append(
            {
            "loss" : loss,
            "y" : y,
            "y_hat" : y_hat
        }
        )
        return {
            "loss" : loss
        }
    
    def _on_second_phase_validation_start(self):
        """Log the training loss and the performance of the second phase
        """
        if len(self.second_phase_step_outputs) > 0:
            train_loss = torch.Tensor([out["loss"] for out in self.second_phase_step_outputs]).detach().mean()
            y = torch.cat([out["y"] for out in self.second_phase_step_outputs if out["y"].numel() != 1]).detach().cpu()
            y_hat = torch.cat([out["y_hat"] for out in self.second_phase_step_outputs if out["y_hat"].numel() != 1]).detach().cpu()
            train_score = self.metric(y_hat, y)
            
            self.log("train_loss", train_loss, prog_bar = True)
            self.log("train_" + self.metric.__name__, train_score, prog_bar = True)
            self.second_phase_step_outputs = []   
            
        return super().on_validation_start()
    
    def _second_phase_validation_epoch_end(self) -> None:
        """Log the validation loss and the performance of the second phase
        """
        val_loss = torch.Tensor([out["loss"] for out in self.second_phase_step_outputs]).mean()

        
        y = torch.cat([out["y"].cpu() for out in self.second_phase_step_outputs if out["y"].numel() != 1])
        y_hat = torch.cat([out["y_hat"].cpu() for out in self.second_phase_step_outputs if out["y_hat"].numel() != 1])
        val_score = self.metric(y_hat, y)

        self.log("val_" + self.metric.__name__, val_score, prog_bar = True)
        self.log("val_loss", val_loss, prog_bar = True)
        self.second_phase_step_outputs = []      
        return super().on_validation_epoch_end()
    
    @abstractmethod
    def predict_step(self, batch: Any, batch_idx: int
    ) -> Any:
        """The perdict step

        Args:
            batch (Any): The input batch
            batch_idx (int): Only for compatibility

        Returns:
            Any: The predicted output and optional additional information.
        """
        pass


from typing import Union, Tuple, Optional, List
from numpy.typing import NDArray

from torch.utils.data import Dataset

import numpy as np
import pandas as pd
import torch
from torch.utils.data import Dataset

from dataclasses import asdict
from ts3l.utils.scarf_utils import SCARFConfig

class SCARFDataset(Dataset):
    def __init__(self, X: pd.DataFrame, 
                        Y: Optional[Union[NDArray[np.int_], NDArray[np.float64]]] = None,
                        unlabeled_data: Optional[pd.DataFrame] = None, 
                        config: Optional[SCARFConfig] = None,
                        continuous_cols: Optional[List] = None, 
                        category_cols: Optional[List] = None,
                        is_regression: Optional[bool] = False,
                        is_second_phase: Optional[bool] = False,
                        ) -> None:
        """A dataset class for SCARF that handles labeled and unlabeled data.

        This class is designed to manage data for the SCARF, accommodating both labeled and unlabeled datasets
        for self-supervised learning scenarios. It supports regression and classification tasks.

        Args:
            X (pd.DataFrame): DataFrame containing the features of the labeled data.
            Y (Union[NDArray[np.int_], NDArray[np.float64]], optional): Numpy array containing the labels for the data. 
                Use integers for classification labels and floats for regression targets. Defaults to None.
            unlabeled_data (pd.DataFrame): DataFrame containing the features of the unlabeled data, used for 
                self-supervised learning. Defaults to None.
            config (SCARFConfig): The given hyperparameter set for SCARF.
            continuous_cols (List, optional): List of continuous columns. Defaults to None.
            category_cols (List, optional): List of categorical columns. Defaults to None.
            is_regression (bool, optional): Flag indicating whether the task is regression (True) or classification (False).
                Defaults to False.
            is_second_phase (bool, optional): Flag indicating whether the dataset is for first phase or second phase learning. 
                Default is False.
        """
        
        if config is not None:
            self.config = config
            
        if unlabeled_data is not None:
            X = pd.concat([X, unlabeled_data])
        
        cat_data = torch.FloatTensor(X[category_cols].values)
        cont_data = torch.FloatTensor(X[continuous_cols].values)
        
        self.data = torch.concat([cat_data, cont_data], dim=1)
        
        self.corruption_rate = self.config.corruption_rate if not is_second_phase else 0.0
        self.corruption_len = int(X.shape[1] * self.corruption_rate)
        if not is_second_phase:
            self.corruption_len = max(1, self.corruption_len)
            
        self.n_sampling_candidate , self.n_features = X.shape

        self.is_regression = is_regression
        
        self.label_class = torch.FloatTensor if is_regression else torch.LongTensor
            
        if Y is None:
            self.label = None
        else:
            self.label = self.label_class(Y)
            
            if self.label_class == torch.LongTensor:
                class_counts = [sum((self.label == i)) for i in set(self.label.numpy())]
                num_samples = len(self.label)

                class_weights = [num_samples/class_counts[i] for i in range(len(class_counts))]
                self.weights = [class_weights[self.label[i]] for i in range(int(num_samples))]


    def __getitem__(self, idx) -> Tuple[torch.Tensor, torch.Tensor]:
        """Retrieves the feature tensor for a given index, along with an label or corrupted version of the feature.

        For first phase learning, this method can return either the original feature tensor or, if corruption is applied, 
        a corrupted version of the feature tensor. For second phase learning, it returns the feature tensor and its corresponding label.

        Args:
            idx (int): The index of the data point to retrieve.

        Returns:
            torch.Tensor: Only the feature tensor for the given index, suitable for test-time inference.
            Tuple[torch.Tensor, torch.Tensor]: A tuple containing the original and corrupted feature tensors for first phase learning, 
                or the feature tensor and its corresponding label for second phase learning,
                or the feature tensor and dummy label for test-time inference.
        """
        if self.label is None:
            if self.corruption_len:
                corruption_mask = torch.zeros((self.n_features), dtype=torch.bool, device=self.data.device)
                corruption_idx = torch.randperm(self.n_features, device=self.data.device)[:self.corruption_len]
                corruption_mask[corruption_idx] = True
                
                # Generate random indices for each feature
                x_random = torch.randint(0, self.n_sampling_candidate, (self.n_features,), device=self.data.device)

                # Use advanced indexing to create the corrupted features
                _x_corrupted = self.data[x_random, torch.arange(self.n_features, device=self.data.device)].float()
                
                x_corrupted = torch.where(corruption_mask, _x_corrupted, self.data[idx])
                
                return self.data[idx], x_corrupted
            return self.data[idx], torch.tensor(-1)
        else:
            return self.data[idx], self.label[idx]

    def __len__(self):
        """Returns the total number of items in the dataset.

        Returns:
            int: The size of the dataset.
        """
        return len(self.data)



import torch
from torch import nn
from typing import OrderedDict, List, Union, Optional

class MLP(nn.Sequential):
    """Simple multi-layer perceptron with activation and optional batch normalization layer and dropout layer"""

    def __init__(self, input_dim: int, hidden_dims: Union[int, List[int]], output_dim = None, n_hiddens: Optional[int] = None, activation: str ="ReLU", use_batch_norm: bool =True, dropout_rate: float =0.0, **kwargs):
        layers = []
        in_dim = input_dim
        
        if use_batch_norm:
            batch_norm_layer = nn.BatchNorm1d # type: ignore
        else:
            batch_norm_layer = nn.Identity # type: ignore
        
        if n_hiddens == None:
            if isinstance(hidden_dims, int):
                n_hiddens = 1
            else:
                n_hiddens = len(hidden_dims)
        
        if isinstance(hidden_dims, int):
            hidden_dims = [hidden_dims for _ in range(n_hiddens)] # type: ignore
        
        if output_dim == None:
            output_dim = hidden_dims[-1]

        for i in range(n_hiddens - 1): # type: ignore
            layers.append(("linear_%d" % i, torch.nn.Linear(in_dim, hidden_dims[i])))
            layers.append(("batchnorm_%d" % i, batch_norm_layer(hidden_dims[i]))) # type: ignore
            layers.append((f"{activation}_%d" % i, getattr(nn, activation)()))
            layers.append(("dropout_%d" % i, torch.nn.Dropout(dropout_rate))) # type: ignore
            in_dim = hidden_dims[i]
        
        if len(hidden_dims) > 0 and len(layers) > 0:
            layers.append((f"linear_{n_hiddens - 1}_layers", torch.nn.Linear(hidden_dims[-1], output_dim))) # type: ignore
        else:
            layers.append((f"linear_{n_hiddens - 1}_layers", torch.nn.Linear(input_dim, output_dim))) # type: ignore
        
        self.output_dim = output_dim
        super().__init__(OrderedDict(layers))


from .mlp_backbone_config import MLPBackboneConfig
from .transformer_config import TransformerBackboneConfig
from .base_backbone_config import BaseBackboneConfig

__all__ = ["MLPBackboneConfig", "TransformerBackboneConfig"]


import torch
from torch import nn
from typing import Union, Type, List

from ts3l.utils import BaseEmbeddingConfig

class FeatureTokenizer(nn.Module):

    def __init__(self, 
                emb_dim : int,
                cont_nums : int,
                cat_cardinality : List[int],
                required_token_dim: int = 1,
                **kwargs,
        ) -> None:
        super().__init__()
        self.emb_dim = emb_dim
        self.cont_nums = cont_nums
        self.cat_nums = len(cat_cardinality)
        
        self.cat_cardinality = cat_cardinality

        bias_dim = 0

        if cont_nums is not None:
            bias_dim += cont_nums

        if cat_cardinality is not None:
            bias_dim += len(cat_cardinality)

            category_offsets = torch.tensor([0] + cat_cardinality[:-1]).cumsum(0)
            self.register_buffer('category_offsets', category_offsets)

            self.cat_weights = nn.Embedding(sum(cat_cardinality), emb_dim)
        
        self.weight = nn.Parameter(torch.Tensor(cont_nums + 1, emb_dim))
        self.bias = nn.Parameter(torch.Tensor(bias_dim, emb_dim))
        
        self.n_features = self.cont_nums + self.cat_nums
        
        if required_token_dim == 2:
            self.forward = self.__generate_tokens
        else:
            self.flatten_dim = self.emb_dim * (self.n_features + 1)
            self.forward = self.__generate_flattened_token

    def __generate_tokens(self, x: torch.Tensor) -> torch.Tensor:
        x_cats, x_conts = x[:, :self.cat_nums].long(), x[:, self.cat_nums:]

        x_conts = torch.cat(
            [torch.ones(
                len(x_conts) if x_conts is not None else len(x_cats), 1, 
                device = x_conts.device if x_conts is not None else x_cats.device)
            ]
            + ([] if x_conts is None else [x_conts]),
            dim = 1
        )

        x = self.weight[None] * x_conts[:, :, None]

        if x_cats is not None:
            x = torch.cat(
                [x, self.cat_weights(x_cats + self.category_offsets[None])],
                dim=1,
            )
        if self.bias is not None:
            bias = torch.cat(
                [
                    torch.zeros(1, self.bias.shape[1], device=x.device),
                    self.bias,
                ]
            )
            x = x + bias[None]

        return x

    def __generate_flattened_token(self, x: torch.Tensor) -> torch.Tensor:

        x = self.__generate_tokens(x)
        
        return x.reshape(-1, self.flatten_dim)
        
class TS3LEmbeddingModule(nn.Module):
    def __init__(self, config: BaseEmbeddingConfig):
        super().__init__()
        self.config = config
        self.input_dim = config.input_dim
        self.output_dim = self.config.output_dim
        self.__set_embedding_layer()
        
    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return self.embeding_layer(x)
    
    def __set_embedding_layer(self):
        if self.config.name == "identity":
            self.embeding_layer = nn.Identity()
        elif self.config.name == "feature_tokenizer":
            self.embeding_layer = FeatureTokenizer(**self.config.__dict__)


import torch
from torch import nn
from torch.nn import functional as F

class NTXentLoss(nn.Module):
    def __init__(self, temperature=1.0):
        """NT-Xent loss for contrastive learning using cosine distance as similarity metric as used in [SimCLR](https://arxiv.org/abs/2002.05709).
        Implementation adapted from https://theaisummer.com/simclr/#simclr-loss-implementation
        Args:
            temperature (float, optional): scaling factor of the similarity metric. Defaults to 1.0.
        """
        super().__init__()
        self.temperature = temperature

    def forward(self, z_i, z_j):
        """Compute NT-Xent loss using only anchor and positive batches of samples. Negative samples are the 2*(N-1) samples in the batch
        Args:
            z_i (torch.tensor): anchor batch of samples
            z_j (torch.tensor): positive batch of samples
        Returns:
            float: loss
        """
        batch_size = z_i.size(0)
        
        # Compute similarity between the sample's embedding and its corrupted view
        z = torch.cat([z_i, z_j], dim=0)
        similarity = F.cosine_similarity(z.unsqueeze(1), z.unsqueeze(0), dim=2)
    
        # Extract positive samples
        positives = similarity[range(batch_size), range(batch_size, 2 * batch_size)]
        positives = torch.cat([positives, similarity[range(batch_size, 2 * batch_size), range(batch_size)]], dim=0)

        # Create mask to exclude self-comparisons
        mask = torch.ones((2 * batch_size, 2 * batch_size), dtype=torch.bool, device=z.device)
        mask.fill_diagonal_(0)

        # Compute numerator and denominator
        exp_sim = torch.exp(similarity / self.temperature)
        numerator = torch.exp(positives / self.temperature)
        denominator = exp_sim * mask

        # Compute loss
        loss = -torch.log(numerator / denominator.sum(dim=1)).mean()
        
        return loss



from dataclasses import dataclass, field
from .base_backbone_config import BaseBackboneConfig
from typing import Optional

@dataclass
class TransformerBackboneConfig(BaseBackboneConfig):
    d_model: Optional[int] = field(default=None)
    ffn_factor: float = field(default=2.0)
    hidden_dim: int = field(default=256)
    encoder_depth: int = field(default=3)
    n_head: int = field(default=2)
    
    def __post_init__(self):
        self.name = "transformer"
        
        if self.d_model is None:
            raise TypeError("__init__ missing 1 required positional argument: 'd_model'")


from typing import Union
from torch.utils.data import Dataset, DataLoader, SequentialSampler, WeightedRandomSampler, RandomSampler

from pytorch_lightning import LightningDataModule

class TS3LDataModule(LightningDataModule):
    """The pytorch lightning datamodule for TabularS3L
    """
    def __init__(self, train_ds:Dataset, val_ds:Dataset, batch_size: int, train_sampler: str, train_collate_fn = None, valid_collate_fn = None, n_jobs: int = 32, drop_last: bool = False, is_regression:bool = False):
        """Initialize the datamodule

        Args:
            train_ds (Dataset): The training dataset.
            val_ds (Dataset): The validation dataset.
            batch_size (int): The batch size of the dataset.
            train_sampler (str): The training sampler for training dataset. It must be one of ["seq", "weighted", "random"].
            train_collate_fn (object): Collate function for train dataloader.
            valid_collate_fn (object): Collate function for validation dataloader.
            n_jobs (int, optional): The number of the cpu core to use. Defaults to 32.
            drop_last (bool, optional): The flag to drop the last batch or not. Defaults to False.
            is_regression (bool, optional): The flag that determines whether the datamodule is for regression task or not. Defaults to False.
        """
        super().__init__()
        self.train_ds = train_ds
        self.val_ds = val_ds

        self.batch_size = batch_size
        
        assert train_sampler in ["seq", "weighted", "random"], 'The train_sampler must be one of the following: "seq", "weighted", or "random".'
        self.train_sampler = train_sampler
        
        self.train_collate_fn = train_collate_fn
        self.valid_collate_fn = valid_collate_fn
        
        self.n_jobs = n_jobs
        self.is_regression = is_regression
        self.drop_last = drop_last

    def setup(self, stage: str):
        """Setup the datamodule for pytorch lightning module of TabularS3L.
        
        Use a weighted random sampler for the second phase step of the classification task, otherwise use a random sampler.
        
        Args:
            stage (str): Only for compatibility, not used
        """
        if self.train_sampler == "seq":
            sampler = SequentialSampler(self.train_ds) # type: ignore
            shuffle = True
        elif self.train_sampler == "weighted":
            sampler = WeightedRandomSampler(self.train_ds.weights, num_samples = len(self.train_ds)) # type: ignore
            shuffle = False
        elif self.train_sampler == "random":
            sampler = RandomSampler(self.train_ds, num_samples = len(self.train_ds)) # type: ignore
            shuffle = False

        self.train_dl = DataLoader(self.train_ds, 
                                    batch_size = self.batch_size, 
                                    shuffle=shuffle, 
                                    sampler = sampler,
                                    num_workers=self.n_jobs,
                                    drop_last=self.drop_last,
                                    collate_fn = self.train_collate_fn)
        self.val_dl = DataLoader(self.val_ds, 
                                batch_size = self.batch_size, 
                                shuffle=False, 
                                sampler = SequentialSampler(self.val_ds), # type: ignore
                                num_workers=self.n_jobs, 
                                drop_last=False, 
                                collate_fn=self.valid_collate_fn)
    
    def train_dataloader(self):
        """Return the training dataloader.
        """
        return self.train_dl

    def val_dataloader(self):
        """Return the validation dataloader.
        """
        return self.val_dl


from dataclasses import dataclass, field
from ts3l.utils import BaseConfig

from typing import Any, List, Optional

@dataclass
class DAEConfig(BaseConfig):
    """ Configuration class for initializing components of the DAELightning Module, including hyperparameters of Denoising AutoEncoder,
    optimizers, learning rate schedulers, and loss functions, along with their respective hyperparameters.

    Inherits Attributes:
        task (str): Specify whether the problem is regression or classification.
        embedding_config (BaseEmbeddingConfig): Configuration for the embedding layer.
        backbone_config (BaseBackboneConfig): Configuration for the backbone network.
        output_dim (int): The dimension of output.
        loss_fn (str): Name of the loss function to be used. Must be an attribute of 'torch.nn'.
        loss_hparams (Dict[str, Any]): Hyperparameters for the loss function. Default is empty dictionary.
        metric (str): Name of the metric to be used. Must be an attribute of 'torchmetrics.functional' or 'sklearn.metrics'. Default is None.
        metric_hparams (Dict[str, Any]): Hyperparameters for the metric. Default is an empty dictionary.
        optim (str): Name of the optimizer to be used. Must be an attribute of 'torch.optim'. Default is 'AdamW'.
        optim_hparams (Dict[str, Any]): Hyperparameters for the optimizer. Default is {'lr': 0.0001, 'weight_decay': 0.00005}.
        scheduler (str): Name of the learning rate scheduler to be used. Must be an attribute of 'torch.optim.lr_scheduler' or None. Default is None.
        scheduler_hparams (Dict[str, Any]): Hyperparameters for the scheduler. Default is None, indicating no scheduler is used.
        initialization (str): The way to initialize neural network parameters. Default is 'kaiming_uniform'.
        random_seed (int): Seed for random number generators to ensure reproducibility. Defaults to 42.
        
    New Attributes:
        noise_type (str): The type of noise to apply. Choices are ["Swap", "Gaussian", "Zero_Out"].
        noise_level (float): Intensity of Gaussian noise to be applied.
        noise_ratio (float): A hyperparameter that is to control the noise ratio during the first phase learning. Default is 0.3.
        mask_loss_weight (float): The special token for unlabeled samples.
        dropout_rate (bool): A hyperparameter that is to control dropout layer. Default is 0.04.
        cat_cardinality (List[int]): The cardinality of categorical features.
        num_continuous (int): The number of continuous features.
        
    Raises:
        ValueError: Inherited from `BaseConfig` to indicate that a configuration for the task, optimizer, scheduler, loss function, or metric is either invalid or not specified.
        ValueError: If the specified 'noise_type' is not in ["Swap", "Gaussian", "Zero_Out"].
        ValueError: If the specified 'noise_level' is not a valid value.
        ValueError: Raised if both `num_categoricals` and `num_continuous` are None, indicating that at least one attribute must be specified.
    """
    
    noise_type: str = field(default="Swap")
    
    noise_level: float = field(default=0)
    
    noise_ratio: float = field(default=0.3)
    
    mask_loss_weight: float = field(default=1.0)
    
    dropout_rate: float = field(default=0.04)
    
    cat_cardinality: List[int] = field(default_factory=lambda: [])
    
    num_continuous: Optional[int] = field(default=None)
    
    def __post_init__(self):
        super().__post_init__()
        
        if self.noise_type not in ["Swap", "Gaussian", "Zero_Out"]:
            raise ValueError('The noise type must be one of ["Swap", "Gaussian", "Zero_Out"], but %s.' % self.noise_type)
        
        if (self.noise_type == "Gaussian") and ((self.noise_level == None) or (self.noise_level <= 0)):
            raise ValueError("The noise level must be a float that is > 0 when the noise type is Gaussian.")
        
        if len(self.cat_cardinality) == 0 and self.num_continuous is None:
            raise ValueError("At least one attribute (num_categorical or num_continuous) must be specified.")
        else:
            self.num_continuous = self.num_continuous if self.num_continuous is not None else 0


from abc import ABC, abstractmethod
import torch
from torch import nn
from typing import Any, Union, Tuple
from ts3l.utils import BaseEmbeddingConfig, BaseBackboneConfig
from .embedding import TS3LEmbeddingModule
from .backbone import TS3LBackboneModule

class TS3LModule(ABC, nn.Module):
    def __init__(self, embedding_config: BaseEmbeddingConfig, backbone_config: BaseBackboneConfig) -> None:
        super(TS3LModule, self).__init__()
        self.embedding_module = TS3LEmbeddingModule(embedding_config)
        self._set_backbone_module(backbone_config)
        
    def __init_subclass__(cls) -> None:
        super().__init_subclass__()
        def init_decorator(cls_init):
            def new_init(self, *args, **kwargs):
                cls_init(self, *args, **kwargs)
                cls.set_first_phase(self)
            return new_init
        cls.__init__ = init_decorator(cls.__init__) # type: ignore
        
    def _set_backbone_module(self, backbone_config):
        self.backbone_module = TS3LBackboneModule(backbone_config)
        
    @property
    @abstractmethod
    def encoder(self):
        raise NotImplementedError
    
    def set_first_phase(self):
        """Set first phase step as the forward pass
        """
        self.forward = self._first_phase_step
        self.encoder.requires_grad_(True)
    
    def set_second_phase(self, freeze_encoder: bool = True):
        """Set second phase step as the forward pass
        """
        self.forward = self._second_phase_step
        self.encoder.requires_grad_(not freeze_encoder)
    
    @abstractmethod
    def _first_phase_step(self, *args: Any, **kwargs: Any) -> Any:
        pass
    
    @abstractmethod
    def _second_phase_step(self, *args: Any, **kwargs: Any) -> Any:
        pass


from typing import Tuple, List
import torch
from torch import nn

def first_phase_step(
    model: nn.Module, batch: Tuple[torch.Tensor, torch.Tensor, torch.Tensor]
) -> Tuple[torch.Tensor, List[torch.Tensor], torch.Tensor]:
    """Forward step of Denoising AutoEncoder during the first phase

    Args:
        model (nn.Module): An instance of Denoising AutoEncoder
        batch (Tuple[torch.Tensor, torch.Tensor, torch.Tensor]): The input batch

    Returns:
        Tuple[torch.Tensor, List[torch.Tensor], torch.Tensor]: The predicted mask vector and the predicted feature vector
    """
    _, x_bar, _ = batch
    mask_preds, cat_preds, cont_preds = model(x_bar)
    return mask_preds, cat_preds, cont_preds


def first_phase_loss(
    x_cat: torch.Tensor,
    x_cont: torch.Tensor,
    mask: torch.Tensor,
    cat_feature_preds: List[torch.Tensor],
    cont_feature_preds: torch.Tensor,
    mask_preds: torch.Tensor,
    mask_loss_fn: nn.Module,
    categorical_loss_fn: nn.Module,
    continuous_loss_fn: nn.Module,
) -> Tuple[torch.Tensor, torch.Tensor]:
    """Calculate the first phase loss of DAE

    Args:
        x_cat (torch.Tensor): The categorical input feature vector
        x_cont (torch.Tensor): The continuous input feature vector
        mask (torch.Tensor): The ground truth mask vector
        cat_feature_preds (List[torch.Tensor]): The predicted categorical feature vector
        cont_feature_preds (torch.Tensor): The predicted continuous feature vector
        mask_preds (torch.Tensor): The predicted mask vector
        mask_loss_fn (nn.Module): The loss function for the mask estimation
        categorical_loss_fn (nn.Module): The loss function for the categorical feature reconstruction
        continuous_loss_fn (nn.Module): The loss function for the continuous feature reconstruction

    Returns:
        Tuple[torch.Tensor, torch.Tensor]: The losses for mask estimation and feature reconstruction
    """
    mask_loss = mask_loss_fn(mask_preds, mask)
    feature_loss = torch.tensor(0.0, device=mask_preds.device)
    
    if x_cat.shape[1] > 0:
        for idx in range(x_cat.shape[1]):
            feature_loss += categorical_loss_fn(cat_feature_preds[idx], x_cat[:, idx].long())
    if x_cont.shape[1] > 0:
        feature_loss += continuous_loss_fn(cont_feature_preds, x_cont)

    return mask_loss, feature_loss


def second_phase_step(
    model: nn.Module, batch: Tuple[torch.Tensor, torch.Tensor]
) -> torch.Tensor:
    """Forward step of Denoising AutoEncoder during the second phase

    Args:
        model (nn.Module): An instance of Denoising AutoEncoder
        batch (Tuple[torch.Tensor, torch.Tensor, torch.Tensor]): The input batch

    Returns:
        torch.Tensor: The predicted label (logit)
    """
    x, _ = batch
    return model(x).squeeze()


def second_phase_loss(
    y: torch.Tensor, y_hat: torch.Tensor, loss_fn: nn.Module
) -> torch.Tensor:
    """Calculate the second phase loss of DAE

    Args:
        y (torch.Tensor): The ground truth label
        y_hat (torch.Tensor): The predicted label
        loss_fn (nn.Module): The loss function for the given task

    Returns:
        torch.Tensor: The loss for the given task
    """
    return loss_fn(y_hat, y)



from dataclasses import dataclass, field

from typing import Dict, Any, Optional, Type
from torch import optim, nn
import torchmetrics
import sklearn

from .embedding_utils import BaseEmbeddingConfig
from .backbone_utils import BaseBackboneConfig

import sys
if sys.version_info >= (3, 8):
    from typing import Literal
else:
    from typing_extensions import Literal

@dataclass
class BaseConfig:
    """ Configuration class for initializing components of the TabularS3L Lightning Module, including optimizers, 
    learning rate schedulers, and loss functions, along with their respective hyperparameters.

    Attributes:
        task (str): Specify whether the problem is regression or classification.
        embedding_config (BaseEmbeddingConfig): Configuration for the embedding layer.
        backbone_config (BaseBackboneConfig): Configuration for the backbone network.
        output_dim (int): The dimension of output.
        loss_fn (str): Name of the loss function to be used. Must be an attribute of 'torch.nn'.
        loss_hparams (Dict[str, Any]): Hyperparameters for the loss function. Default is empty dictionary.
        metric (str): Name of the metric to be used. Must be an attribute of 'torchmetrics.functional' or 'sklearn.metrics'. Default is None.
        metric_hparams (Dict[str, Any]): Hyperparameters for the metric. Default is an empty dictionary.
        optim (str): Name of the optimizer to be used. Must be an attribute of 'torch.optim'. Default is 'AdamW'.
        optim_hparams (Dict[str, Any]): Hyperparameters for the optimizer. Default is {'lr': 0.0001, 'weight_decay': 0.00005}.
        scheduler (str): Name of the learning rate scheduler to be used. Must be an attribute of 'torch.optim.lr_scheduler' or None. Default is None.
        scheduler_hparams (Dict[str, Any]): Hyperparameters for the scheduler. Default is None, indicating no scheduler is used.
        initialization (str): The way to initialize neural network parameters. Default is 'kaiming_uniform'.
        random_seed (int): Seed for random number generators to ensure reproducibility. Defaults to 42.
    Raises:
        ValueError: If the specified 'optim' is not a valid optimizer in 'torch.optim'.
        ValueError: If the specified 'scheduler' is not None and is not a valid scheduler in 'torch.optim.lr_scheduler'.
        ValueError: If the specified 'loss_fn' is not None and is not a valid loss function in 'torch.nn'.        
        ValueError: If the specified 'metric' is not a valid metric in 'torchmetrics' or 'sklearn.metrics'.
        ValueError: If the specified 'task' is not a valid task in ['regression', 'classification']'.
        ValueError: If the combination of `backbone_config` and `embedding_config` is invalid.
    """
    task: str
    
    embedding_config: BaseEmbeddingConfig
    
    backbone_config: BaseBackboneConfig
    
    output_dim: int
    
    loss_fn: str
    
    loss_hparams: Dict[str, Any] = field(default_factory=dict)
    
    metric: Optional[str] = field(default=None)
    
    metric_hparams: Dict[str, Any] = field(default_factory=dict)
    
    optim: str = field(default="AdamW")
    
    optim_hparams: Dict[str, Any] = field(
                                            default_factory=lambda: {
                                                    "lr" : 0.0001,
                                                    "weight_decay" : 0.00005
                                            }
                                        )
    
    scheduler: Optional[str] = field(default=None)
    
    scheduler_hparams: Optional[Dict[str, Any]] = field(default = None)
    
    initialization: Literal['xavier_uniform', 'xavier_normal', 'kaiming_uniform', 'kaiming_normal', 'uniform', 'normal'] = "kaiming_uniform"
    
    random_seed: int = field(default=42)
    
    
    def __post_init__(self):

        if (type(self.task) is not str or (self.task != "regression" and self.task != "classification")):
            raise ValueError(f"{self.task} is not a valid task. Choices are: ['regression', 'classification']")
        
        if type(self.optim) is not str or not hasattr(optim, self.optim):
            raise ValueError(f"{self.optim} is not a valid optimizer in torch.optim")
        
        if self.scheduler is not None and (type(self.scheduler) is not str or not hasattr(optim.lr_scheduler, self.scheduler)):
            raise ValueError(f"{self.scheduler} is not a valid scheduler in torch.optim.lr_scheduler")

        if type(self.loss_fn) is not str or not hasattr(nn, self.loss_fn):
            raise ValueError(f"{self.loss_fn} is not a valid loss function in torch.nn")
        
        if self.metric is None:
            if self.task == "regression":
                self.metric = "mean_squared_error"
            else:
                self.metric = "accuracy_score"
                
        elif (type(self.metric) is not str or (not hasattr(torchmetrics.functional, self.metric) and not hasattr(sklearn.metrics, self.metric))):
            raise ValueError(f"{self.metric} is not a valid metric in torchmetrics.functional or sklearn.metrics")
        
        if self.backbone_config.name == "transformer" and self.embedding_config.name == "identity":
            raise ValueError(f"Transformer backbone and Identity Embedding are not compatible.")
        
        if self.backbone_config.name == "transformer" and self.embedding_config.required_token_dim == 1:
            raise ValueError(f"Embedding's 'required_token_dim' should be 2 for transformer backbone, not {self.embedding_config.required_token_dim}")


from .base_embedding_config import BaseEmbeddingConfig
from dataclasses import dataclass, field
from typing import List

@dataclass
class IdentityEmbeddingConfig(BaseEmbeddingConfig):
    def __post_init__(self):
        self.name = "identity"
        self.output_dim = self.input_dim


import torch
from torch import nn

from typing import Tuple, List, Union
from ts3l.models.common import TS3LModule
from ts3l.utils import BaseEmbeddingConfig, BaseBackboneConfig

class Projector(nn.Module):
    def __init__(self, hidden_dim: int) -> None:
        """Initializes the projector module used in the SwitchTab

        Args:
            hidden_dim (int): The dimensionality of both the input and output of the projector.
        """
        super().__init__()
        
        self.linear = nn.Linear(hidden_dim, hidden_dim)
        self.activation = nn.SiLU()
        
    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return self.linear(self.activation(x))

class Decoder(nn.Module):
    def __init__(self, hidden_dim: int, output_dim:int) -> None:
        """Initializes the decoder module used in the SwitchTab

        Args:
            hidden_dim (int): The dimensionality of the output of the projector.
            output_dim (int): The dimensionality of input features of SwitchTab.
        """
        super().__init__()
        
        self.linear = nn.Linear(hidden_dim * 2, output_dim)
        self.activation = nn.SiLU()
        
    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return self.linear(self.activation(x))

class SwitchTab(TS3LModule):
    def __init__(self,
                    embedding_config: BaseEmbeddingConfig,
                    backbone_config: BaseBackboneConfig,
                    output_dim: int,
                    **kwargs) -> None:
        """Initialize SwitchTab

        Args:
            embedding_config (BaseEmbeddingConfig): Configuration for the embedding layer.
            backbone_config (BaseBackboneConfig): Configuration for the backbone network.
            output_dim (int): The dimensionality of the output.
        """
        super(SwitchTab, self).__init__(embedding_config, backbone_config)
        self.output_dim = output_dim
        self.__return_salient_feature = False
        
        self.projector_m = Projector(self.backbone_module.output_dim)
        self.projector_s = Projector(self.backbone_module.output_dim)
        
        self.decoder = Decoder(self.backbone_module.output_dim, self.embedding_module.input_dim)
        self.head = nn.Linear(self.backbone_module.output_dim, output_dim)
        self.activation = nn.SiLU()

    @property
    def encoder(self) -> nn.Module:
        return self.backbone_module
        
    @property
    def return_salient_feature(self) -> bool:
        """Gets the value of the private attribute '__return_salient_feature' which indicates whether 
        salient features should be returned by the model.

        Returns:
            bool: The current state of the '__return_salient_feature' flag.
        """
        return self.__return_salient_feature
    
    @return_salient_feature.setter
    def return_salient_feature(self, flag: bool) -> None:
        """Sets the value of the private attribute '__return_salient_feature' to control whether 
        salient features should be returned by the model.

        Args:
            flag (bool): A boolean value to set the '__return_salient_feature' attribute.
        """
        self.__return_salient_feature = flag

    def _first_phase_step(self, x: torch.Tensor) -> Tuple[torch.Tensor, torch.Tensor]:
        """The first phase step of SwitchTab
        Processes the given samples to decuple salient and mutual embeddings across data samples.
        
        Args:
            x (torch.Tensor): The input batch

        Returns:
            Tuple[torch.Tensor, torch.Tensor]: Reconstructed tensors and predicted labels.
        """
        
        size = len(x) // 2
        x = self.embedding_module(x)
        zs = self.encoder(x)
        
        ms = self.projector_m(zs)
        ss = self.projector_s(zs)
        
        m1, s1 = ms[:size], ss[:size]
        m2, s2 = ms[size:], ss[size:]
        
        x1_tilde_hat = torch.concat([torch.concat([m1, s1], dim=1), torch.concat([m2, s1], dim=1)])
        x2_hat_tilde = torch.concat([torch.concat([m1, s2], dim=1), torch.concat([m2, s2], dim=1)])

        x1_recover_switch = self.decoder(x1_tilde_hat)
        x2_switch_recover = self.decoder(x2_hat_tilde)
        
        x_hat = torch.concat([x1_recover_switch, x2_switch_recover])
        y_hat = self.head(self.activation(zs))

        return x_hat, y_hat

    
    def _second_phase_step(self, 
                x : torch.Tensor) -> Union[torch.Tensor, Tuple[torch.Tensor, torch.Tensor]]:
        """The second phase step of SwitchTab

        Args:
            x (torch.Tensor): The input batch

        Returns:
            Union[torch.Tensor, Tuple[torch.Tensor, torch.Tensor]]: If 'return_salient_feature' is False,
                returns a tensor 'y_hat' representing the predicted label. If True, returns a tuple
                where the first element is 'y_hat' and the second element is a tensor of salient features extracted
                from the input.
        """
        x = self.embedding_module(x)
        emb = self.encoder(x)
        y_hat = self.head(self.activation(emb))
        if not self.return_salient_feature:
            return y_hat
        else:
            salient_features = self.projector_s(emb)
            return y_hat, salient_features


from typing import Dict, Any, Tuple
import torch

from .base_module import TS3LLightining
from ts3l.models import SCARF
from ts3l.models.scarf import NTXentLoss
from ts3l.utils.scarf_utils import SCARFConfig
from ts3l import functional as F
from ts3l.utils import BaseConfig

class SCARFLightning(TS3LLightining):

    def __init__(self, config: SCARFConfig) -> None:
        """Initialize the pytorch lightining module of SCARF

        Args:
            config (SubTabConfig): The configuration of SCARFLightning.
        """
        super(SCARFLightning, self).__init__(config)

    def _initialize(self, config: BaseConfig):
        """Initializes the model with specific hyperparameters and sets up various components of SCARFLightning.

        Args:
            config (Dict[str, Any]): The given hyperparameter set for SCARF.
        """
        if not isinstance(config, SCARFConfig):
            raise TypeError(f"Expected SCARFConfig, got {type(config)}")
        
        self.contrastive_loss = NTXentLoss(config.tau)

        self._init_model(SCARF, config)

    def _get_first_phase_loss(self, batch: Tuple[torch.Tensor, torch.Tensor]) -> torch.Tensor:
        """Calculate the first phase loss

        Args:
            batch (Tuple[torch.Tensor, torch.Tensor]): The input batch

        Returns:
            torch.Tensor: The final loss of first phase step
        """

        emb_anchor, emb_corrupted = F.scarf.first_phase_step(self.model, batch)

        loss = F.scarf.first_phase_loss(
            emb_anchor, emb_corrupted, self.contrastive_loss
        )

        return loss

    def _get_second_phase_loss(
        self, batch: Tuple[torch.Tensor, torch.Tensor]
    ) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
        """Calculate the second phase loss

        Args:
            batch (Tuple[torch.Tensor, torch.Tensor]): The input batch

        Returns:
            torch.Tensor: The final loss of second phase step
            torch.Tensor: The label of the labeled data
            torch.Tensor: The predicted label of the labeled data
        """
        _, y = batch
        y_hat = F.scarf.second_phase_step(self.model, batch)
        task_loss = F.scarf.second_phase_loss(y, y_hat, self.task_loss_fn)

        return task_loss, y, y_hat

    def set_second_phase(self, freeze_encoder: bool = False) -> None:
        """Set the module to fine-tuning
        
        Args:
            freeze_encoder (bool): If True, the encoder will be frozen during fine-tuning. Otherwise, the encoder will be trainable.
                                    Default is False.
        """
        return super().set_second_phase(freeze_encoder)
    
    def predict_step(self, batch: Tuple[torch.Tensor, torch.Tensor], batch_idx: int) -> torch.Tensor:
        """The perdict step of SCARF

        Args:
            batch (Tuple[torch.Tensor, torch.Tensor]): The input batch
            batch_idx (int): For compatibility, do not use

        Returns:
            torch.Tensor: The predicted output (logit)
        """

        y_hat = F.scarf.second_phase_step(self.model, batch)

        return y_hat




import pandas as pd
from typing import Any, List, Union, Optional, Tuple
from numpy.typing import NDArray

import torch
from torch.utils.data import Dataset

import numpy as np
from ts3l.utils.switchtab_utils import SwitchTabConfig

class SwitchTabDataset(Dataset):
    def __init__(self,
                X: pd.DataFrame,
                Y: Optional[Union[NDArray[np.int_], NDArray[np.float64]]] = None, 
                config: Optional[SwitchTabConfig] = None, 
                unlabeled_data: Optional[pd.DataFrame] = None, 
                u_label: Optional[Any] = -1,
                continuous_cols: Optional[List] = None, 
                category_cols: Optional[List] = None, 
                is_second_phase: Optional[bool] = False,
                is_regression: Optional[bool] = False, 
        ) -> None:
        """A dataset class for SwitchTab that handles labeled and unlabeled data.

        This class is designed to manage data for the SwitchTab, accommodating both labeled and unlabeled datasets
        for self- and semi-supervised learning scenarios. It supports regression and classification tasks.

        Args:
            X (pd.DataFrame): DataFrame containing the features of the labeled data.
            Y (Union[NDArray[np.int_], NDArray[np.float64]], optional): Numpy array containing the labels for the data. 
                Use integers for classification labels and floats for regression targets. Defaults to None.
            config (SwitchTabConfig): The given hyperparameter set for SwitchTab.
            unlabeled_data (pd.DataFrame): DataFrame containing the features of the unlabeled data, used for 
                self-supervised learning. Defaults to None.
            u_label (int, optional): The specifier for unlabeled sample. Defaults to -1.
            continuous_cols (List, optional): List of continuous columns. Defaults to None.
            category_cols (List, optional): List of categorical columns. Defaults to None.
            is_second_phase (bool): The flag that determines whether the dataset is for first phase or second phase learning. Default is False.
            is_regression (bool, optional): Flag indicating whether the task is regression (True) or classification (False).
                Defaults to False.
        """
        
        if config is not None:
            self.config = config
            
        if unlabeled_data is not None:
            X = pd.concat([X, unlabeled_data])
        
        cat_data = torch.FloatTensor(X[category_cols].values)
        cont_data = torch.FloatTensor(X[continuous_cols].values)
        
        self.data = torch.concat([cat_data, cont_data], dim=1)
        
        self.u_label = u_label
        self.label = None
        self.label_class = torch.FloatTensor if is_regression else torch.LongTensor
        
        if not is_second_phase:
            self.corruption_rate = self.config.corruption_rate if not is_second_phase else 0.0
            self.corruption_len = int(X.shape[1] * self.corruption_rate)
            self.n_sampling_candidate, self.n_features = X.shape
            
        if Y is not None:
            
            self.label = self.label_class(Y)
            
            if unlabeled_data is not None:
                self.label = torch.concat((self.label, self.label_class([u_label for _ in range(len(unlabeled_data))])), dim=0) # type: ignore
            
            if not is_regression:
                class_counts = [sum((self.label == i)) for i in set(self.label.numpy())]
                num_samples = len(self.label)

                class_weights = [num_samples/class_counts[i] for i in range(len(class_counts))]
                self.weights = [class_weights[self.label[i]] for i in range(int(num_samples))]
            else:
                self.weights = [1.0 for _ in range(len(X))]
                if unlabeled_data is not None:
                    unlabeled_weight = len(X) / len(unlabeled_data)
                    self.weights.extend([unlabeled_weight for _ in range(len(unlabeled_data))])
        

        if not is_second_phase:
            self.__getitem = self.__first_phase_get_item # type: ignore
        else:
            self.__getitem = self.__second_phase_get_item # type: ignore

    
    def __getitem__(self, idx: int) -> Union[Tuple[torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor], Tuple[torch.Tensor, torch.Tensor], torch.Tensor]:
        """Return the feature tensor for a given index, along with an optional label or corrupted version of the tensor.

        For first phase learning, this method returns the original and a corrupted feature tensor pair with optional labels.
        For second phase learning, it returns the feature tensor and its corresponding label.

        Args:
            idx (int): The index of the data point to retrieve.

        Returns:
            Tuple[torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor]: A pair of original and corrupted feature tensors with optional labels.
            Tuple[torch.Tensor, torch.Tensor]: A feature tensor and its corresponding label for second phase learning.
        """
        return self.__getitem(idx)
    
    def __first_phase_get_item(self, idx: int) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor]:
        """Return the original and a corrupted feature tensor pair with optional labels for the first phase learning

        Args:
            idx (int): The index of the data to sample

        Returns:
            Tuple[torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor]: A pair of original and corrupted feature tensors with optional labels.
        """
        x_1 = self.data[idx]
        
        idx2 = idx
        while idx2 == idx:
            idx2 = np.random.randint(0, len(self.data))
        x_2 = self.data[idx2]
        
        if self.label is not None:
            y_1, y_2 = self.label[idx], self.label[idx2] # type: ignore
        else:
            y_1, y_2 = self.label_class((self.u_label,)), self.label_class((self.u_label,))
        
        if self.corruption_len:
            xc_1 = self.__generate_corrupted_sample(x_1)
            xc_2 = self.__generate_corrupted_sample(x_2)
            return x_1, xc_1, y_1, x_2, xc_2, y_2
        else:
            return x_1, x_1, y_1, x_2, x_2, y_2

    def __second_phase_get_item(self, idx) -> Tuple[torch.Tensor, torch.Tensor]:
        """Return a input and label pair for the second phase learning

        Args:
            idx (int): The index of the data to sample

        Returns:
            Tuple[torch.Tensor, torch.Tensor]: A tuple containing the feature tensor and corresponding label.
        """
        if self.label is not None:
            return self.data[idx], self.label[idx]
        else:
            return self.data[idx], torch.tensor(-1)
            
    def __len__(self):
        """Return the length of the dataset
        """
        return len(self.data)
    
    def __generate_corrupted_sample(self, x: torch.Tensor) -> torch.Tensor:
        """Return a corrupted version of the feature tensor for a given sample.

        Args:
            x (torch.Tensor): A feature tensor to be corrupted.

        Returns:
            torch.Tensor: A corrupted feature tensor.
        """
        corruption_mask = torch.zeros((self.n_features), dtype=torch.bool, device=x.device)
        corruption_idx = torch.randperm(self.n_features, device=x.device)[:self.corruption_len]
        corruption_mask[corruption_idx] = True
        
        # Generate random indices for each feature
        x_random = torch.randint(0, self.n_sampling_candidate, (self.n_features,), device=x.device)

        # Use advanced indexing to create the corrupted features
        _x_corrupted = self.data[x_random, torch.arange(self.n_features, device=x.device)].float()
        
        x_corrupted = torch.where(corruption_mask, _x_corrupted, x)
        return x_corrupted
    
class SwitchTabFirstPhaseCollateFN(object):
    """A callable class designed for batch processing, specifically tailored for the first phase learning with SwitchTab. 
    It restructures the batch by concatenating certain elements to form new tensors.
    This class is meant to be used as a collate function in a DataLoader, where it efficiently organizes batch data 
    for training during first phase learning.
    """
    def __call__(self, batch: Tuple) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
        """Process the input batch to generate tensors by concatenating specific elements.

        Args:
            batch (Tuple): The batch to process. The structure of each element is expected to be 
                (x1, x1_corrupted, y1, x2, x2_corrupted, y2).

        Returns:
            Tuple[torch.Tensor, torch.Tensor, torch.Tensor]: A tuple containing three tensors:
                - xs (Tensor): Concatenation of the x1 and x2 from each tuple in the batch.
                - xcs (Tensor): Concatenation of the x1_corrupted and x2_corrupted.
                - ys (Tensor): Concatenation of the y1 and y2.
        """
        xs = torch.cat([torch.stack([x1 for x1, _, _, _, _, _ in batch]), torch.stack([x2 for _, _, _, x2, _, _ in batch])])
        xcs = torch.cat([torch.stack([x1c for _, x1c, _, _, _, _ in batch]), torch.stack([x2c for _, _, _, _, x2c, _ in batch])])
        ys = torch.cat([torch.stack([y1 for _, _, y1, _, _, _ in batch]), torch.stack([y2 for _, _, _, _, _, y2 in batch])])
        return xs, xcs, ys


from typing import Dict, Any, Tuple
import torch
from torch import nn

from .base_module import TS3LLightining
from ts3l.models import VIME
from ts3l.utils.vime_utils import VIMEConfig
from ts3l import functional as F
from ts3l.utils import BaseConfig

class VIMELightning(TS3LLightining):
    
    def __init__(self, config: VIMEConfig) -> None:
        """Initialize the pytorch lightining module of VIME

        Args:
            config (VIMEConfig): The configuration of VIMELightning.
        """
        super(VIMELightning, self).__init__(config)
    
    def _initialize(self, config: BaseConfig) -> None:
        """Initializes the model with specific hyperparameters and sets up various components of VIMELightning.

        Args:
            config (Dict[str, Any]): The given hyperparameter set for VIME. 
        """
        
        if not isinstance(config, VIMEConfig):
            raise TypeError(f"Expected VIMEConfig, got {type(config)}")
        
        self.alpha1 = config.alpha1
        self.alpha2 = config.alpha2
        
        self.beta = config.beta
        
        self.K = config.K
        
        self.num_categoricals, self.num_continuous = len(config.cat_cardinality), config.num_continuous
        
        self.u_label = config.u_label
        
        self.mask_loss_fn = nn.BCELoss()
        self.categorical_feature_loss_fn = nn.CrossEntropyLoss()
        self.continuous_feature_loss_fn = nn.MSELoss()
        
        self.consistency_loss_fn = nn.MSELoss()

        self._init_model(VIME, config)
    
    def _get_first_phase_loss(self, batch: Tuple[torch.Tensor, torch.Tensor, torch.Tensor]):
        """Calculate the first phase loss

        Args:
            batch (Tuple[torch.Tensor, torch.Tensor, torch.Tensor]): The input batch

        Returns:
            torch.FloatTensor: The final loss of first phase step
        """
        
        mask_preds, cat_preds, cont_preds = F.vime.first_phase_step(self.model, batch)
        
        mask_loss, categorical_feature_loss, continuous_feature_loss = F.vime.first_phase_loss(
            batch[2][:, : self.num_categoricals],
            batch[2][:, self.num_categoricals :],
            batch[1],
            cat_preds,
            cont_preds,
            mask_preds,
            self.mask_loss_fn,
            self.categorical_feature_loss_fn,
            self.continuous_feature_loss_fn,
        )
        
        return mask_loss + self.alpha1 * categorical_feature_loss + self.alpha2 * continuous_feature_loss
    
    def _get_second_phase_loss(self, batch: Tuple[torch.Tensor, torch.Tensor]):
        """Calculate the second phase loss

        Args:
            batch (Tuple[torch.Tensor, torch.Tensor]): The input batch

        Returns:
            torch.FloatTensor: The final loss of second phase step
            torch.Tensor: The label of the labeled data
            torch.Tensor: The predicted label of the labeled data
        """
        _, y = batch

        y_hat = F.vime.second_phase_step(self.model, batch)
        
        labeled_idx = (y != self.u_label).flatten()
        unlabeled_idx = (y == self.u_label).flatten()
        
        labeled_y_hat = y_hat[labeled_idx]
        labeled_y = y[labeled_idx].squeeze()
        
        unlabeled_y_hat = y_hat[unlabeled_idx]
        
        task_loss, consistency_loss = F.vime.second_phase_loss(labeled_y, labeled_y_hat, unlabeled_y_hat, self.consistency_loss_fn, self.task_loss_fn, self.K)
        
        loss = task_loss + self.beta * consistency_loss
        
        return loss, labeled_y, labeled_y_hat
    
    def set_second_phase(self, freeze_encoder: bool = True) -> None:
        """Set the module to fine-tuning
        
        Args:
            freeze_encoder (bool): If True, the encoder will be frozen during fine-tuning. Otherwise, the encoder will be trainable.
                                    Default is True.
        """
        return super().set_second_phase(freeze_encoder)
    
    def predict_step(self, batch, batch_idx: int
        ) -> torch.FloatTensor:
        """The predict step of VIME

        Args:
            batch (Dict[str, Any]): The input batch
            batch_idx (int): Only for compatibility, do not use

        Returns:
            torch.FloatTensor: The predicted output (logit)
        """
        y_hat = self(batch[0])

        return y_hat


from typing import Dict, Any, Tuple, Union, Optional, List
from numpy.typing import NDArray

import torch
from torch.utils.data import Dataset
import numpy as np
import pandas as pd

from dataclasses import asdict
from ts3l.utils.dae_utils import DAEConfig

class DAEDataset(Dataset):
    def __init__(self, X: pd.DataFrame, 
                        Y: Optional[Union[NDArray[np.int_], NDArray[np.float64]]] = None,
                        unlabeled_data: Optional[pd.DataFrame] = None,
                        continuous_cols: Optional[List] = None, 
                        category_cols: Optional[List] = None,
                        is_regression: Optional[bool] = False
                        ) -> None:
        """A dataset class for DenoisingAutoEncoder that handles labeled and unlabeled data.

        This class is designed to manage data for the DenoisingAutoEncoder, accommodating both labeled and unlabeled datasets
        for self-supervised learning scenarios. It supports regression and classification tasks.

        Args:
            X (pd.DataFrame): DataFrame containing the features of the labeled data.
            Y (Optional[Union[NDArray[np.int_], NDArray[np.float64]]]): Numpy array containing the labels for the data. 
                Use integers for classification labels and floats for regression targets. Defaults to None.
            unlabeled_data (Optional[pd.DataFrame]): DataFrame containing the features of the unlabeled data, used for 
                self-supervised learning. Defaults to None.
            continuous_cols (List, optional): List of continuous columns. Defaults to None.
            category_cols (List, optional): List of categorical columns. Defaults to None.
            is_regression (Optional[bool]): Flag indicating whether the task is regression (True) or classification (False).
                Defaults to False.
        """
        
        if unlabeled_data is not None:
            X = pd.concat([X, unlabeled_data])
        
        self.len = len(X)
            
        self.cont_data = torch.FloatTensor(X[continuous_cols].values)
        self.cat_data = torch.FloatTensor(X[category_cols].values)
        
        self.continuous_cols = continuous_cols
        self.category_cols = category_cols
        
        self.label_class = torch.FloatTensor if is_regression else torch.LongTensor
            
        if Y is None:
            self.label = None
        else:
            self.label = self.label_class(Y)
            
            if not is_regression:
                class_counts = [sum((self.label == i)) for i in set(self.label.numpy())]
                num_samples = len(self.label)

                class_weights = [num_samples/class_counts[i] for i in range(len(class_counts))]
                self.weights = [class_weights[self.label[i]] for i in range(int(num_samples))]
    
    def __getitem__(self, idx: int) -> Union[torch.Tensor, Tuple[torch.Tensor, torch.Tensor]]:
        """Retrieves the feature and label tensors for a given index.

        Args:
            idx (int): The index of the data point to retrieve.

        Returns:
            Union[torch.Tensor, Tuple[torch.Tensor, torch.Tensor]]: The feature tensor and the label tensor (if available) for the given index.
        """
        
        cat_samples = self.cat_data[idx]

        cont_samples = self.cont_data[idx]

        x = torch.concat((cat_samples, cont_samples))
        
        if self.label is None:
            return x
        
        return x, self.label[idx]
    
    def __len__(self) -> int:
        """Returns the total number of items in the dataset.

        Returns:
            int: The size of the dataset.
        """
        return self.len
    
class DAECollateFN(object):
    def __init__(self, config: DAEConfig) -> None:
        """A collate function for DenoisingAutoEncoder to generate noisy x for dataloaders.

        Args:
            config (SubTabConfig): The given hyperparameter set for DenoisingAutoEncoder.
        """
        
        self.config = config
        
        self.noise_ratio = config.noise_ratio
        self.noise_type = config.noise_type
        self.noise_level = config.noise_level
    
    def __generate_noisy_xbar(self, x : torch.Tensor) -> torch.Tensor:
        """Generates a noisy version of the input sample `x`.

        Args:
            x (torch.Tensor): The original sample.

        Returns:
            torch.Tensor: The noisy sample.
        """
        no, dim = x.shape
        
        # Initialize corruption array
        x_bar = torch.zeros([no, dim]).to(x.device)

        # Randomly (and column-wise) shuffle data
        if self.noise_type == "Swap":
            # Generate random permutations for all columns
            permutations = torch.stack([torch.randperm(no, device=x.device) for _ in range(dim)], dim=1)
            # Use advanced indexing to permute the tensor
            x_bar = x[permutations, torch.arange(dim, device=x.device)]
            
        elif self.noise_type == "Gaussian":
            noise = torch.normal(mean=0.0, std=self.noise_level, size=x.shape, device=x.device)
            x_bar = x + noise

        return x_bar
    
    def __generate_x_tilde(self, x: torch.Tensor) -> Tuple[torch.Tensor, torch.Tensor]:
        """Generates noisy samples for the given x.

        Args:
            x (torch.Tensor): The original sample.

        Returns:
            torch.Tensor: The noisy samples.
        """
        x_bar_noisy = self.__generate_noisy_xbar(x)
        
        mask = torch.bernoulli(torch.full(x.shape, self.noise_ratio, device=x.device))
        
        x_bar = x * (1 - mask) + x_bar_noisy * mask
        
        return x_bar, mask
    
    def __call__(self, batch: Tuple[torch.Tensor, ...]) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
        """Generates noisy x from the given input x.

        Args:
            batch (Tuple[torch.Tensor, ...]): A batch of the original samples.

        Returns:
            Tuple[torch.Tensor, torch.Tensor]: A tuple containing the original samples, and the noisy samples.
        """

        x = torch.stack(batch)

        x_bar, mask = self.__generate_x_tilde(x)
        
        return x, x_bar, mask


from .base_embedding_config import BaseEmbeddingConfig
from dataclasses import dataclass, field
from typing import List

@dataclass
class FTEmbeddingConfig(BaseEmbeddingConfig):
    emb_dim: int = field(default=128)
    cont_nums: int = field(default=0)
    cat_cardinality: List[int] = field(default_factory=lambda: [])
    required_token_dim: int = field(default=1)
    
    def __set_output_dim(self):
        if self.required_token_dim == 1:
            self.output_dim = self.emb_dim * (self.cont_nums + len(self.cat_cardinality) + 1)
        else:
            self.output_dim = self.cont_nums + len(self.cat_cardinality)
            
    def __post_init__(self):
        self.name = "feature_tokenizer"
        self.__set_output_dim()
        if not self.required_token_dim in [1, 2]:
            raise ValueError(f"{self.required_token_dim} is not a valid value. Choices are: [1, 2]")


import torch
import torch.nn as nn

class VIMEPredictor(nn.Module):
    def __init__(self, input_dim:int, hidden_dim:int, output_dim:int):
        """_summary_

        Args:
            input_dim (int): The input dimension of the predictor. Must be same to the dimension of the encoder.
            hidden_dim (int): The hidden dimension of the predictor
            output_dim (int): The output dimension of the predictor
        """
        super().__init__()
        self.fc1 = nn.Linear(input_dim, hidden_dim)
        self.fc2 = nn.Linear(hidden_dim, hidden_dim)
        self.fc3 = nn.Linear(hidden_dim, output_dim)

    def forward(self, x):
        """The forward pass of semi-supervised module of VIME

        Args:
            x (torch.FloatTensor): The input batch.

        Returns:
            torch.FloatTensor: The predicted logits of VIME
        """
        x = torch.relu(self.fc1(x))
        x = torch.relu(self.fc2(x))
        logits = self.fc3(x)
        return logits



from typing import Dict, Any, Type, Tuple, Union
import torch
from torch import nn

from .base_module import TS3LLightining
from ts3l.models import SwitchTab
from ts3l.utils.switchtab_utils import SwitchTabConfig
from ts3l import functional as F
from ts3l.utils import BaseConfig

class SwitchTabLightning(TS3LLightining):
    
    def __init__(self, config: SwitchTabConfig) -> None:
        """Initialize the pytorch lightining module of SwitchTab

        Args:
            config (SwitchTabConfig): The configuration of SwitchTabLightning.
        """
        super(SwitchTabLightning, self).__init__(config)
    
    def _initialize(self, config: BaseConfig) -> None:
        """Initializes the model with specific hyperparameters and sets up various components of SwitchTabLightning.

        Args:
            config (Dict[str, Any]): The given hyperparameter set for SwitchTab. 
        """
        if not isinstance(config, SwitchTabConfig):
            raise TypeError(f"Expected SwitchTabConfig, got {type(config)}")
        
        self.u_label = config.u_label
        self.alpha = config.alpha
        
        self.reconstruction_loss_fn = nn.MSELoss()
        
        self._init_model(SwitchTab, config)
    
    def _get_first_phase_loss(self, batch: Tuple[torch.Tensor, torch.Tensor, torch.Tensor]):
        """Calculate the first phase loss

        Args:
            batch (Tuple[torch.Tensor, torch.Tensor, torch.Tensor]): The input batch

        Returns:
            torch.FloatTensor: The final loss of first phase step
        """
        
        x_hat, labeled_y, labeled_y_hat = F.switchtab.first_phase_step(self.model, batch, self.u_label)

        xs, _, _ = batch
        
        size = len(batch)
        
        xs = torch.concat(
        [torch.concat([xs[:size], xs[:size]]), torch.concat([xs[size:], xs[size:]])]
    )
        
        recon_loss, task_loss = F.switchtab.first_phase_loss(xs, x_hat, labeled_y, labeled_y_hat, self.reconstruction_loss_fn, self.task_loss_fn)
    
        return recon_loss + self.alpha * task_loss
    
    def _get_second_phase_loss(self, batch: Tuple[torch.Tensor, torch.Tensor]) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
        """Calculate the second phase loss

        Args:
            batch (Tuple[torch.Tensor, torch.Tensor]: The input batch

        Returns:
            torch.FloatTensor: The final loss of second phase step
            torch.Tensor: The label of the labeled data
            torch.Tensor: The predicted label of the labeled data
        """
        _, y = batch
        y_hat = F.switchtab.second_phase_step(self.model, batch)
        task_loss = F.switchtab.second_phase_loss(y, y_hat, self.task_loss_fn)
        
        return task_loss, y, y_hat
    
    def set_second_phase(self, freeze_encoder: bool = False) -> None:
        """Set the module to fine-tuning
        
        Args:
            freeze_encoder (bool): If True, the encoder will be frozen during fine-tuning. Otherwise, the encoder will be trainable.
                                    Default is False.
        """
        return super().set_second_phase(freeze_encoder)
    
    def predict_step(self, batch: Tuple[torch.Tensor, torch.Tensor], batch_idx: int
        ) -> Union[torch.Tensor, Tuple[torch.Tensor, torch.Tensor]]:
        """The predict step of SwitchTab

        Args:
            batch (torch.Tensor): The input batch
            batch_idx (int): Only for compatibility, do not use

        Returns:
            Union[torch.Tensor, Tuple[torch.Tensor, torch.Tensor]]: The predicted output (logit) or (logit, salient_feature)
        """
        y_hat = F.switchtab.second_phase_step(self.model, batch)

        return y_hat
        
    def return_salient_feature(self, flag: bool) -> None:
        """Configures the model to either return or not return salient features based on the provided flag.

        Args:
            flag (bool): A boolean flag that determines the behavior of returning salient features.
                        If True, the model will include salient features in its output; if False, it will not.
        """
        self.model.return_salient_feature(flag)


import torch
from torch import nn
import numpy as np
import itertools
from torch.nn import functional as F
class JointLoss(nn.Module):
    """ JointLoss for SubTab during the first phase learning.
    
    When computing the constrastive loss, we use a similarity matrix of size (N x k) x N. 
    he matrix includes k positive samples and all other samples are considered negatives. 
    The matrix is shown below as an 8x4 array, assuming a batch size of 4 and 2 subsets.
                                                        P . . .
                                                        P . . .
                                                        . P . .
                                                        . P . .
                                                        . . P .
                                                        . . P .
                                                        . . . P
                                                        . . . P
    """

    def __init__(self,
                 tau: float,
                 n_subsets: int,
                 use_contrastive: bool = True,
                 use_distance: bool = True,
                 use_cosine_similarity: bool = False
        ) -> None:
        super(JointLoss, self).__init__()


        self.n_subsets = n_subsets
        
        # Temperature to use scale logits
        self.temperature = tau

        self.use_cosine_similarity = use_cosine_similarity
        
        # Function to generate similarity matrix: Cosine, or Dot product
        self.similarity_fn = self._cosine_simililarity if use_cosine_similarity else self._dot_simililarity
        
        # Loss function
        self.mse_loss = nn.MSELoss()
        self.criterion = nn.CrossEntropyLoss(reduction="sum")
        
        self.use_contrastive = use_contrastive
        self.use_distance = use_distance
        

    @staticmethod
    def _dot_simililarity(x, y):
        # Reshape x: (2N, C) -> (2N, 1, C)
        x = x.unsqueeze(1)
        # Reshape y: (2N, C) -> (1, C, 2N)
        y = y.T.unsqueeze(0)
        # Similarity shape: (2N, 2N)
        similarity = torch.tensordot(x, y, dims=2)
        return similarity

    def _cosine_simililarity(self, x, y):
        # Reshape x: (2N, C) -> (2N, 1, C)
        x = x.unsqueeze(1)
        # Reshape y: (2N, C) -> (1, C, 2N)
        y = y.unsqueeze(0)
        # Similarity shape: (2N, 2N)
        return F.cosine_similarity(x, y, dim=-1)
    
    def get_anchor_loss(self, similarity: torch.Tensor) -> torch.Tensor:
        batch_size = similarity.size(0)
        group_size = self.n_subsets
        
        # Create a mask to exclude self-similarity
        identity_mask = torch.eye(
            batch_size, dtype=torch.bool, device=similarity.device
        )

        # Create masks for positive and negative pairs
        group_indices = torch.arange(batch_size, device=similarity.device) // group_size
        group_mask = group_indices.unsqueeze(0) == group_indices.unsqueeze(1)

        positives_mask = group_mask & ~identity_mask
        negatives_mask = ~group_mask

        # Compute positive and negative sums
        pos_sum = torch.sum(torch.exp(similarity) * positives_mask.float(), dim=1)
        neg_sum = torch.sum(torch.exp(similarity) * negatives_mask.float(), dim=1)

        # Exclude zero positive sums to avoid log(0)
        pos_sum = torch.clamp(pos_sum, min=1e-10)

        # Compute anchor losses
        anchor_loss = -torch.log(pos_sum / (pos_sum + neg_sum))

        return anchor_loss
        
    def XNegloss(self, projections: torch.FloatTensor) -> torch.Tensor:
        
        # Compute cosine similarity using the provided function
        similarity = self.similarity_fn(projections, projections)
        
        # return torch.stack([self.get_anchor_loss(turn, similarity[turn]) for turn in range(len(similarity))]).mean()
        anchor_losses = self.get_anchor_loss(similarity)
        
        return anchor_losses.mean()

    def forward(self, projections, xrecon, xorig):
        """
        Args:
            projections (torch.FloatTensor): Projections for each subset.
            xrecon (torch.FloatTensor): Reconstructed sample x from subsets.
            xorig (torch.FloatTensor): Original features of x
        """
        
        # recontruction loss
        recon_loss = self.mse_loss(xrecon, xorig)

        # Initialize contrastive and distance losses with recon_loss as placeholder
        closs, dist_loss = None, None

        # Start with default loss i.e. reconstruction loss
        loss = recon_loss

        if self.use_contrastive:
            closs = self.XNegloss(projections)
            loss = loss + closs

        if self.use_distance:
            # recontruction loss for z
            combi = np.array(list(itertools.combinations(range(self.n_subsets), 2)))
            left = combi[:, 0]
            right = combi[:, 1]
            
            # Create an index tensor for the pairs
            indices = torch.arange(len(projections)).view(-1, self.n_subsets)
            left_indices = indices[:, left].reshape(-1)
            right_indices = indices[:, right].reshape(-1)

            # Compute the MSE loss in a vectorized manner
            dist_loss = self.mse_loss(projections[left_indices], projections[right_indices])
        
            loss += dist_loss

        return loss, closs, recon_loss, dist_loss



from .vime_lightning import VIMELightning
from .subtab_lightning import SubTabLightning
from .scarf_lightning import SCARFLightning
from .dae_lightning import DAELightning
from .switchtab_lightning import SwitchTabLightning

__all__ = ["VIMELightning", "SubTabLightning", "SCARFLightning", "DAELightning"]


from typing import Tuple, List

import torch
import torch.nn as nn

from ts3l.models.common import TS3LModule
from ts3l.models.common.reconstruction_head import ReconstructionHead
from .vime_predictor import VIMEPredictor

from ts3l.utils import BaseEmbeddingConfig, BaseBackboneConfig

class VIME(TS3LModule):
    def __init__(self, 
                embedding_config: BaseEmbeddingConfig,  backbone_config: BaseBackboneConfig, num_continuous: int, cat_cardinality: List[int], predictor_dim: int, output_dim: int, **kwargs):
        """Initialize VIME

        Args:
            embedding_config (BaseEmbeddingConfig): Configuration for the embedding layer.
            backbone_config (BaseBackboneConfig): Configuration for the backbone network.
            num_continuous (int): The number of continuous features.
            cat_cardinality (List[int]): The cardinality of categorical features.
            predictor_dim (int): The hidden dimension of the predictor.
            output_dim (int): The output dimension of the predictor.
        """
        super(VIME, self).__init__(embedding_config, backbone_config)

        self.mask_predictor = nn.Linear(self.backbone_module.output_dim, embedding_config.input_dim, bias=True)
        self.feature_predictor = ReconstructionHead(self.backbone_module.output_dim, num_continuous, cat_cardinality)
        self.predictor = VIMEPredictor(self.backbone_module.output_dim, predictor_dim, output_dim)
        
    @property
    def encoder(self) -> nn.Module:
        return self.backbone_module
        
    def _first_phase_step(self, x: torch.Tensor) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
        """The first phase step of VIME

        Args:
            x (torch.Tensor): The input batch

        Returns:
            Tuple[torch.Tensor, torch.Tensor, torch.Tensor]: The predicted mask vector and predicted features
        """
        x = self.embedding_module(x)
        x = self.encoder(x)
        mask_output, (cat_preds, cont_preds) = torch.sigmoid(self.mask_predictor(x)), self.feature_predictor(x)
        return mask_output, cat_preds, cont_preds
    
    
    def _second_phase_step(self, x: torch.Tensor) -> torch.Tensor:
        """The second phase step of VIME

        Args:
            x (torch.Tensor): The input batch.

        Returns:
            torch.Tensor: The predicted logits of VIME
        """
        x = self.embedding_module(x)
        x = self.encoder(x)
        logits = self.predictor(x)
        return logits


from dataclasses import dataclass, field
from ts3l.utils import BaseConfig

@dataclass
class SubTabConfig(BaseConfig):
    """ Configuration class for initializing components of the SubTabLightning Module, including hyperparameters of SubTab,
    optimizers, learning rate schedulers, and loss functions, along with their respective hyperparameters.

    Inherits Attributes:
        task (str): Specify whether the problem is regression or classification.
        embedding_config (BaseEmbeddingConfig): Configuration for the embedding layer.
        backbone_config (BaseBackboneConfig): Configuration for the backbone network.
        output_dim (int): The dimension of output.
        loss_fn (str): Name of the loss function to be used. Must be an attribute of 'torch.nn'.
        loss_hparams (Dict[str, Any]): Hyperparameters for the loss function. Default is empty dictionary.
        metric (str): Name of the metric to be used. Must be an attribute of 'torchmetrics.functional' or 'sklearn.metrics'. Default is None.
        metric_hparams (Dict[str, Any]): Hyperparameters for the metric. Default is an empty dictionary.
        optim (str): Name of the optimizer to be used. Must be an attribute of 'torch.optim'. Default is 'AdamW'.
        optim_hparams (Dict[str, Any]): Hyperparameters for the optimizer. Default is {'lr': 0.0001, 'weight_decay': 0.00005}.
        scheduler (str): Name of the learning rate scheduler to be used. Must be an attribute of 'torch.optim.lr_scheduler' or None. Default is None.
        scheduler_hparams (Dict[str, Any]): Hyperparameters for the scheduler. Default is None, indicating no scheduler is used.
        initialization (str): The way to initialize neural network parameters. Default is 'kaiming_uniform'.
        random_seed (int): Seed for random number generators to ensure reproducibility. Defaults to 42.
        
    New Attributes:
        projection_dim (int): The dimension of projection module. Default is 256.
        tau (float): A hyperparameter that is to scale similarity between projections during the first phase.
        use_cosine_similarity (bool):  A hyperparameter that is to select whether using cosine similarity or dot similarity when calculating similarity
                                        between projections during the first phase. Default is False.
        use_contrastive (bool): A hyperparameter that is to select using contrastive loss or not during the first phase. Default is True.
        use_distance (bool): A hyperparameter that is to select using distance loss or not during the first phase. Default is True.
        n_subsets (int): The number of subsets to generate different views of the data. Default is 4.
        overlap_ratio (float): A hyperparameter that is to control the extent of overlapping between the subsets. Default is 0.75.
        shuffle (bool): Whether to shuffle the subsets. 
        mask_ratio (float): Ratio of features to be masked as noise.
        noise_type (str): The type of noise to apply.
        noise_level (float): Intensity of Gaussian noise to be applied.

    Raises:
        ValueError: Inherited from `BaseConfig` to indicate that a configuration for the task, optimizer, scheduler, loss function, or metric is either invalid or not specified.
        ValueError: If the specified 'noise_type' is not in ["Swap", "Gaussian", "Zero_Out"].
        ValueError: If the specified 'noise_level' is not a valid value.
    """
    
    projection_dim: int = field(default=256)
    
    tau: float = field(default=0.1)
    
    use_cosine_similarity: bool = field(default=False)
    
    use_contrastive: bool = field(default=True)
    
    use_distance: bool = field(default=True)
    
    n_subsets: int = field(default=4)
    
    overlap_ratio: float = field(default=0.75)
    
    shuffle: bool = field(default=False)
    
    mask_ratio: float = field(default=0.2)
    
    noise_type: str = field(default="Swap")
    
    noise_level: float = field(default=0)
    
    def __post_init__(self):
        super().__post_init__()
        
        if self.noise_type not in ["Swap", "Gaussian", "Zero_Out"]:
            raise ValueError('The noise type must be one of ["Swap", "Gaussian", "Zero_Out"], but %s.' % self.noise_type)
        
        if (self.noise_type == "Gaussian") and (self.noise_level <= 0):
            raise ValueError("The noise level must be a float that is > 0 when the noise type is Gaussian.")


from typing import Tuple
import torch
from torch import nn

def first_phase_step(
    model: nn.Module, batch: Tuple[torch.Tensor, torch.Tensor]
) -> Tuple[torch.Tensor, torch.Tensor]:
    """Forward step of SCARF during the first phase

    Args:
        model (nn.Module): An instance of SCARF.
        batch (Tuple[torch.Tensor, torch.Tensor]): The input batch.

    Returns:
        Tuple[torch.Tensor, torch.Tensor]: The anchor vector and the corrupted vector.
    """
    x, x_corrupted = batch
    emb_anchor, emb_corrupted = model(x, x_corrupted)
    return emb_anchor, emb_corrupted    


def first_phase_loss(
    emb_anchor: torch.Tensor, 
    emb_corrupted: torch.Tensor,
    contrastive_loss_fn: nn.Module
) -> torch.Tensor:
    """Calculate the first phase loss of SCARF

    Args:
        emb_anchor (torch.Tensor): The anchor vector.
        emb_corrupted (torch.Tensor): The corrupted vector.
        contrastive_loss_fn (nn.Module): The loss function for the contrastive learning.

    Returns:
        torch.Tensor: The contrastive loss between the anchor vector and the corrupted vector.
    """
    loss = contrastive_loss_fn(emb_anchor, emb_corrupted)
    return loss


def second_phase_step(
    model: nn.Module, batch: Tuple[torch.Tensor, torch.Tensor]
) -> torch.Tensor:
    """Forward step of SCARF during the second phase.

    Args:
        model (nn.Module): An instance of SCARF.
        batch (Tuple[torch.Tensor, torch.Tensor]): The input batch.

    Returns:
        torch.Tensor: The predicted label (logit).
    """
    x, _ = batch
    y_hat = model(x).squeeze()
    return y_hat


def second_phase_loss(
    y: torch.Tensor, y_hat: torch.Tensor, task_loss_fn: nn.Module
) -> torch.Tensor:
    """Calculate the second phase loss of SCARF.

    Args:
        y (torch.Tensor): The ground truth label.
        y_hat (torch.Tensor): The predicted label.
        task_loss_fn (nn.Module): The loss function for the given task.

    Returns:
        torch.Tensor: The loss for the given task.
    """
    return task_loss_fn(y_hat, y)



from typing import Dict, Any, Tuple, Union, Type

from .base_module import TS3LLightining
from ts3l.models import SubTab
from ts3l.models.subtab import JointLoss

import torch
from ts3l.utils.subtab_utils import SubTabConfig
from ts3l import functional as F
from ts3l.utils import BaseConfig
class SubTabLightning(TS3LLightining):
    
    def __init__(self, config: SubTabConfig) -> None:
        """Initialize the pytorch lightining module of SubTab

        Args:
            config (SubTabConfig): The configuration of SubTabLightning.
        """
        super(SubTabLightning, self).__init__(config)

    def _initialize(self, config: BaseConfig):
        """Initializes the model with specific hyperparameters and sets up various components of SubTabLightning.

        Args:
            config (Dict[str, Any]): The given hyperparameter set for SubTab. 
        """
        if not isinstance(config, SubTabConfig):
            raise TypeError(f"Expected SubTabConfig, got {type(config)}")
        
        self.joint_loss_fn = JointLoss(
                                        config.tau,
                                        config.n_subsets,
                                        config.use_contrastive,
                                        config.use_distance,
                                        use_cosine_similarity = config.use_cosine_similarity
                                        )

        self.n_subsets = config.n_subsets
        
        self._init_model(SubTab, config)

    def _get_first_phase_loss(self, batch:Tuple[torch.Tensor, torch.Tensor]) -> torch.Tensor:
        """Calculate the first phase loss

        Args:
            batch (Tuple[torch.Tensor, torch.Tensor, torch.Tensor]): The input batch

        Returns:
            torch.Tensor: The final loss of first phase step
        """
        projections, x_recons = F.subtab.first_phase_step(self.model, batch)
        
        x_originals, _ = batch
        
        total_loss, contrastive_loss, recon_loss, dist_loss = F.subtab.first_phase_loss(projections, x_recons, x_originals, self.n_subsets, self.joint_loss_fn)
        
        return total_loss
    
    def _get_second_phase_loss(self, batch: Tuple[torch.Tensor, torch.Tensor]):
        """Calculate the second phase loss

        Args:
            batch (Tuple[torch.Tensor, torch.Tensor, torch.Tensor]): The input batch

        Returns:
            torch.FloatTensor: The final loss of second phase step
            torch.Tensor: The label of the labeled data
            torch.Tensor: The predicted label of the labeled data
        """
        y_hat = F.subtab.second_phase_step(self.model, batch)
        
        _, y = batch
        
        task_loss = F.subtab.second_phase_loss(y, y_hat, self.task_loss_fn)
        
        return task_loss, y, y_hat
    
    def set_second_phase(self, freeze_encoder: bool = True) -> None:
        """Set the module to fine-tuning
        
        Args:
            freeze_encoder (bool): If True, the encoder will be frozen during fine-tuning. Otherwise, the encoder will be trainable.
                                    Default is True.
        """
        return super().set_second_phase(freeze_encoder)
    
    def predict_step(self, batch: Tuple[torch.Tensor, torch.Tensor], batch_idx: int) -> torch.Tensor:
        """The perdict step of SubTab

        Args:
            batch (Tuple[torch.Tensor, torch.Tensor, torch.Tensor]: The input batch
            batch_idx (int): For compatibility, do not use

        Returns:
            torch.FloatTensor: The predicted output (logit)
        """
        
        return F.subtab.second_phase_step(self.model, batch)


from .base_embedding_config import BaseEmbeddingConfig
from .feature_tokenizer_config import FTEmbeddingConfig
from .identity_config import IdentityEmbeddingConfig

__all__ = ["BaseEmbeddingConfig", "FTEmbeddingConfig", "IdentityEmbeddingConfig"]


from typing import Tuple, Dict, Any
import torch
from torch import nn

def get_duplicated_input(x: torch.Tensor, n_subsets: int) -> torch.Tensor:
    """Duplicates the input tensor across the batch dimension to match the number of subsets for reconstruction loss.

    Args:
        x (torch.Tensor): The input tensor to duplicate.
        n_subsets (int): The number of subsets for each sample.

    Returns:
        torch.Tensor: A tensor with the input 'n_subsets' times along the batch dimension.
    """
    x_duplicated = x
    for _ in range(1, n_subsets):
        x_duplicated = torch.concat((x_duplicated, x), dim = 0)
    return x_duplicated

def arrange_tensors(x: torch.Tensor, n_subsets: int) -> torch.Tensor:
    """
    Rearranges the input tensor into a sequence of subsets concatenated along the first dimension.

    This function reshapes the input tensor 'x' of shape '(no, dim)' into a tensor of shape 
    '(n_subsets, samples, dim)', where 'no' is the total number of tensors, 'dim' is the feature 
    dimension, and 'samples' is the batch size (calculated as 'no // n_subsets'). 
    The reshaped tensor is then split into subsets, which are concatenated along the first dimension to produce 
    the output tensor.

    Args:
        projections (torch.Tensor): A tensor of shape '(no, dim)' where 'no' is the total number of tensors and 'dim' is the feature dimension. 
        n_subsets (int): The number of subsets for each sample.

    Returns:
        torch.Tensor: A reshaped tensor where subsets of tensors are concatenated along the first dimension. 
                        The resulting shape is '(no, dim)', maintaining the original dimensionality 
                        but rearranging the order of observations to align with subset divisions.
    """
    no, dim = x.shape
    samples = int(no / n_subsets)
    
    x = x.reshape((n_subsets, samples, dim))
    
    return torch.concat([x[:, i] for i in range(samples)])
    
def first_phase_step(
    model: nn.Module, batch: Tuple[torch.Tensor, torch.Tensor]
) -> Tuple[torch.Tensor, torch.Tensor]:
    """Forward step of SubTab during the first phase.

    Args:
        model (nn.Module): An instance of SubTab.
        batch (Tuple[torch.Tensor, torch.Tensor, torch.Tensor]): The input batch.

    Returns:
        Tuple[torch.Tensor, torch.Tensor]: The projections of each subset and the reconstructed input feature vectors.
    """

    x, _ = batch
    
    projections, x_recons = model(x)
    
    return projections, x_recons


def first_phase_loss(
    projections: torch.Tensor, 
    x_recons: torch.Tensor, 
    x_originals: torch.Tensor, 
    n_subsets: int, 
    joint_loss_fn: nn.Module
) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor]:
    """Calculate the first phase loss of SubTab.

    Args:
        projections (torch.Tensor): The projections of each subset.
        x_recons (torch.Tensor): The reconstructed input tensor from the each subset.
        x_originals (torch.Tensor): The original input tensors for the reconstruction loss.
        n_subsets (int): The number of subsets for each sample.
        joint_loss_fn (nn.Module): The joint loss function for SubTab during the first phase learning.

    Returns:
        Tuple[torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor]: The total loss, contrastive loss, reconstruction loss, 
                                                                        and distance loss during the first phase of learning.
    """
    
    x_originals = get_duplicated_input(x_originals, n_subsets)
    
    arranged_projections = arrange_tensors(projections, n_subsets)
    
    total_loss, contrastive_loss, recon_loss, dist_loss = joint_loss_fn(arranged_projections, x_recons, x_originals)
    
    return total_loss, contrastive_loss, recon_loss, dist_loss


def second_phase_step(
    model: nn.Module, batch: Tuple[torch.Tensor, torch.Tensor]
) -> torch.Tensor:
    """Forward step of SubTab during the second phase.

    Args:
        model (nn.Module): An instance of SubTab.
        batch (Tuple[torch.Tensor, torch.Tensor, torch.Tensor]): The input batch.

    Returns:
        torch.Tensor: The predicted label (logit).
    """
    x, _ = batch
    return model(x).squeeze()

def second_phase_loss(
    y: torch.Tensor,
    y_hat: torch.Tensor,
    task_loss_fn: nn.Module,
) -> torch.Tensor:
    """Calculate the second phase loss of SubTab

    Args:
        y (torch.Tensor): The ground truth label.
        y_hat (torch.Tensor): The predicted label.
        task_loss_fn (nn.Module): The loss function for the given task.
    Returns:
        torch.Tensor: The losse for the given task.
    """

    task_loss = task_loss_fn(y_hat, y)
    
    return task_loss



from typing import Tuple, Any
import torch
from torch import nn


def first_phase_step(
    model: nn.Module,
    batch: Tuple[torch.Tensor, torch.Tensor, torch.Tensor],
    u_label: Any = -1,
) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
    """Forward step of SwitchTab during the first phase.

    Args:
        model (nn.Module): An instance of SwitchTab.
        batch (Tuple[torch.Tensor, torch.Tensor, torch.Tensor]): The input batch.
        u_label (Any): The specifier for unlabeled sample. Defaults to -1.

    Returns:
        Tuple[torch.Tensor, torch.Tensor, torch.Tensor]: The reconstructed input vector, the ground truth label vector, and the predicted label vector for the labeled samples.
    """

    _, xcs, ys = batch
    
    cls_idx = ys != u_label

    x_hat, y_hat = model(xcs)

    labeled_y_hat, labeled_y = y_hat[cls_idx.squeeze(), :].squeeze(), ys[cls_idx]

    return x_hat, labeled_y, labeled_y_hat


def first_phase_loss(
    xs: torch.Tensor,
    x_hat: torch.Tensor,
    labeled_y: torch.Tensor,
    labeled_y_hat: torch.Tensor,
    reconstruction_loss_fn: nn.Module,
    task_loss_fn: nn.Module,
) -> Tuple[torch.Tensor, torch.Tensor]:
    """Calculate the first phase loss of SwitchTab

    Args:
        xs (torch.Tensor): The input vector.
        x_hat (torch.Tensor): The reconstructed input vector.
        labeled_y (torch.Tensor): The ground truth labels.
        labeled_y_hat (torch.Tensor): The predicted labels for the labeled samples.
        reconstruction_loss_fn (nn.Module): The loss function for the feature reconstruction.
        task_loss_fn (nn.Module): The loss function for the given task.

    Returns:
        Tuple[torch.Tensor, torch.Tensor]: The reconstruction loss and the task loss.
    """

    recon_loss = reconstruction_loss_fn(x_hat, xs)

    task_loss = torch.tensor(0, device=xs.device)
    if len(labeled_y) > 0:
        task_loss = task_loss_fn(labeled_y_hat, labeled_y)

    return recon_loss, task_loss


def second_phase_step(
    model: nn.Module, batch: Tuple[torch.Tensor, torch.Tensor]
) -> torch.Tensor:
    """Forward step of SwitchTab during the second phase.

    Args:
        model (nn.Module): An instance of SwitchTab.
        batch (Tuple[torch.Tensor, torch.Tensor]): The input batch.

    Returns:
        torch.Tensor: The predicted label (logit).
    """
    x, _ = batch
    y_hat = model(x).squeeze()
    return y_hat


def second_phase_loss(
    y: torch.Tensor, y_hat: torch.Tensor, task_loss_fn: nn.Module
) -> torch.Tensor:
    """Calculate the second phase loss of SwitchTab.

    Args:
        y (torch.Tensor): The ground truth label.
        y_hat (torch.Tensor): The predicted label.
        task_loss_fn (nn.Module): The loss function for the given task.

    Returns:
        torch.Tensor: The loss for the given task.
    """
    return task_loss_fn(y_hat, y)



from typing import Dict, Any, Type, Tuple
import torch
from torch import nn

from .base_module import TS3LLightining
from ts3l.models import DAE
from ts3l.utils.dae_utils import DAEConfig
from ts3l import functional as F
from ts3l.utils import BaseConfig

class DAELightning(TS3LLightining):

    def __init__(self, config: DAEConfig) -> None:
        """Initialize the pytorch lightining module of DAE

        Args:
            config (DAEConfig): The configuration of DAELightning.
        """
        super(DAELightning, self).__init__(config)

    def _initialize(self, config: BaseConfig) -> None:
        """Initializes the model with specific hyperparameters and sets up various components of DAELightning.

        Args:
            config (Dict[str, Any]): The given hyperparameter set for DAE.
        """
        if not isinstance(config, DAEConfig):
            raise TypeError(f"Expected DAEConfig, got {type(config)}")
        
        self.mask_loss_weight = config.mask_loss_weight

        self.num_categoricals, self.num_continuous = (
            len(config.cat_cardinality),
            config.num_continuous,
        )

        self.mask_loss_fn = nn.BCELoss()
        self.categorical_feature_loss = nn.CrossEntropyLoss()
        self.continuous_feature_loss = nn.MSELoss()

        self._init_model(DAE, config)

    def _get_first_phase_loss(
        self, batch: Tuple[torch.Tensor, torch.Tensor, torch.Tensor]
    ):
        """Calculate the first phase loss

        Args:
            batch (Tuple[torch.Tensor, torch.Tensor, torch.Tensor]): The input batch

        Returns:
            torch.FloatTensor: The final loss of first phase step
        """
        x, _, mask = batch
        
        mask_preds, cat_preds, cont_preds = F.dae.first_phase_step(self.model, batch)
        
        mask_loss, feature_loss = F.dae.first_phase_loss(
            x[:, : self.num_categoricals],
            x[:, self.num_categoricals :],
            mask,
            cat_preds, 
            cont_preds,
            mask_preds,
            self.mask_loss_fn,
            self.categorical_feature_loss,
            self.continuous_feature_loss,
        )
        
        return mask_loss * self.mask_loss_weight + feature_loss

    def _get_second_phase_loss(self, batch: Tuple[torch.Tensor, torch.Tensor]):
        """Calculate the second phase loss

        Args:
            batch (Tuple[torch.Tensor, torch.Tensor]): The input batch

        Returns:
            torch.FloatTensor: The final loss of second phase step
            torch.Tensor: The ground truth label
            torch.Tensor: The predicted label
        """
        _, y = batch

        y_hat = F.dae.second_phase_step(self.model, batch)

        loss = F.dae.second_phase_loss(y, y_hat, self.task_loss_fn)

        return loss, y, y_hat

    def set_second_phase(self, freeze_encoder: bool = True) -> None:
        """Set the module to fine-tuning
        
        Args:
            freeze_encoder (bool): If True, the encoder will be frozen during fine-tuning. Otherwise, the encoder will be trainable.
                                    Default is True.
        """
        return super().set_second_phase(freeze_encoder)
    
    def predict_step(self, batch, batch_idx: int) -> torch.FloatTensor:
        """The predict step of DAE

        Args:
            batch (torch.Tensor): The input batch
            batch_idx (int): Only for compatibility, do not use

        Returns:
            torch.FloatTensor: The predicted output (logit)
        """
        y_hat = self(batch)

        return y_hat



import torch.nn.init as init

def initialize_weights(model, initialization='kaiming_normal', exclude_layers=["norm"]):
    """
    Initializes the weights and biases of the given PyTorch model using the specified initialization method.
    
    Args:
        model (nn.Module): PyTorch model instance.
        initialization (str): Name of the initialization method.
            Options: 'xavier_uniform', 'xavier_normal', 'kaiming_uniform', 'kaiming_normal', 'uniform', 'normal'.
            Defaults to 'kaiming_normal'.
    """

    for name, param in model.named_parameters():
        if any(layer_name.upper() in name.upper() for layer_name in exclude_layers):
            continue
        if 'weight' in name:
            if initialization == 'xavier_uniform':
                init.xavier_uniform_(param)
            elif initialization == 'xavier_normal':
                init.xavier_normal_(param)
            elif initialization == 'kaiming_uniform':
                init.kaiming_uniform_(param)
            elif initialization == 'kaiming_normal':
                init.kaiming_normal_(param)
            elif initialization == 'uniform':
                init.uniform_(param, -0.1, 0.1)  # Adjust range as needed
            elif initialization == 'normal':
                init.normal_(param, mean=0, std=0.01)  # Adjust mean and std as needed
            else:
                raise ValueError("Invalid initialization method. Please choose from 'xavier_uniform', 'xavier_normal', 'kaiming_uniform', 'kaiming_normal', 'uniform', or 'normal'.")
        elif 'bias' in name:
            init.constant_(param, 0.0)  # Initialize biases to zero


from dataclasses import dataclass, field
from ts3l.utils import BaseConfig

from typing import Any, List, Optional

@dataclass
class VIMEConfig(BaseConfig):
    """ Configuration class for initializing components of the VIMELightning Module, including hyperparameters of VIME,
    optimizers, learning rate schedulers, and loss functions, along with their respective hyperparameters.

    Inherits Attributes:
        task (str): Specify whether the problem is regression or classification.
        embedding_config (BaseEmbeddingConfig): Configuration for the embedding layer.
        backbone_config (BaseBackboneConfig): Configuration for the backbone network.
        output_dim (int): The dimension of output.
        loss_fn (str): Name of the loss function to be used. Must be an attribute of 'torch.nn'.
        loss_hparams (Dict[str, Any]): Hyperparameters for the loss function. Default is empty dictionary.
        metric (str): Name of the metric to be used. Must be an attribute of 'torchmetrics.functional' or 'sklearn.metrics'. Default is None.
        metric_hparams (Dict[str, Any]): Hyperparameters for the metric. Default is an empty dictionary.
        optim (str): Name of the optimizer to be used. Must be an attribute of 'torch.optim'. Default is 'AdamW'.
        optim_hparams (Dict[str, Any]): Hyperparameters for the optimizer. Default is {'lr': 0.0001, 'weight_decay': 0.00005}.
        scheduler (str): Name of the learning rate scheduler to be used. Must be an attribute of 'torch.optim.lr_scheduler' or None. Default is None.
        scheduler_hparams (Dict[str, Any]): Hyperparameters for the scheduler. Default is None, indicating no scheduler is used.
        initialization (str): The way to initialize neural network parameters. Default is 'kaiming_uniform'.
        random_seed (int): Seed for random number generators to ensure reproducibility. Defaults to 42.
        
    New Attributes:
        predictor_dim (int): The hidden dimension of predictor. Default is 256.
        cat_cardinality (List[int]): The cardinality of categorical features.
        num_continuous (int): The number of continuous features.
        u_label (Any): The special token for unlabeled samples.
        alpha1 (float): A hyperparameter that is to control the trade-off between 
                        the mask estimation and categorical feature estimation loss during first phase. 
                        Default is 2.0.
        alpha2 (float): A hyperparameter that is to control the trade-off between 
                        the mask estimation and continuous feature estimation loss during first phase. 
                        Default is 2.0.
        beta (float): A hyperparameter that is to control the trade-off between 
                        the supervised and unsupervised loss during second phase. 
                        Default is 1.0.
        K (int): The number of augmented samples for consistency regularization. Default is 3.
        p_m (float): A hyperparameter that is to control the masking ratio during the first phase learning. Default is 0.3.

    Raises:
        ValueError: Inherited from `BaseConfig` to indicate that a configuration for the task, optimizer, scheduler, loss function, or metric is either invalid or not specified.
        ValueError: Raised if both `num_categoricals` and `num_continuous` are None, indicating that at least one attribute must be specified.
        
    """
    
    predictor_dim: int = field(default=256)
    
    # num_categoricals: Optional[int] = field(default=None)
    cat_cardinality: List[int] = field(default_factory=lambda: [])
    
    num_continuous: Optional[int] = field(default=None)
    
    u_label: Any = field(default=-1)
    
    alpha1: float = field(default=2.0)
    
    alpha2: float = field(default=2.0)
    
    beta: float = field(default=1.0)
    
    K: int = field(default=3)
    
    p_m: float = field(default=0.3)
    
    def __post_init__(self):
        super().__post_init__()
        
        if len(self.cat_cardinality) == 0 and self.num_continuous is None:
            raise ValueError("At least one attribute (num_categorical or num_continuous) must be specified.")
        else:
            self.num_continuous = self.num_continuous if self.num_continuous is not None else 0


from typing import Tuple, List
import torch
from torch import nn


def first_phase_step(
    model: nn.Module, batch: Tuple[torch.Tensor, torch.Tensor, torch.Tensor]
) -> Tuple[torch.Tensor, List[torch.Tensor], torch.Tensor]:
    """Forward step of VIME during the first phase

    Args:
        model (nn.Module): An instance of VIME
        batch (Tuple[torch.Tensor, torch.Tensor, torch.Tensor]): The input batch

    Returns:
        Tuple[torch.Tensor, List[torch.Tensor], torch.Tensor]: The predicted mask vector and the predicted feature vector
    """
    mask_preds, cat_preds, cont_preds = model(batch[0])
    return mask_preds, cat_preds, cont_preds


def first_phase_loss(
    x_cat: torch.Tensor,
    x_cont: torch.Tensor,
    mask: torch.Tensor,
    cat_feature_preds: List[torch.Tensor],
    cont_feature_preds: torch.Tensor,
    mask_preds: torch.Tensor,
    mask_loss_fn: nn.Module,
    categorical_loss_fn: nn.Module,
    continuous_loss_fn: nn.Module,
) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
    """Calculate the first phase loss of VIME

    Args:
        x_cat (torch.Tensor): The categorical input feature vector
        x_cont (torch.Tensor): The continuous input feature vector
        mask (torch.Tensor): The ground truth mask vector
        cat_feature_preds (torch.Tensor): The predicted categorical feature vector
        cont_feature_preds (torch.Tensor): The predicted continuous feature vector
        mask_preds (torch.Tensor): The predicted mask vector
        mask_loss_fn (nn.Module): The loss function for the mask estimation
        categorical_loss_fn (nn.Module): The loss function for the categorical feature reconstruction
        continuous_loss_fn (nn.Module): The loss function for the continuous feature reconstruction

    Returns:
        Tuple[torch.Tensor, torch.Tensor, torch.Tensor]: The losses for mask estimation and feature reconstruction
    """
    mask_loss = mask_loss_fn(mask_preds, mask)
    categorical_feature_loss = torch.tensor(0.0, device=mask_preds.device)
    continuous_feature_loss = torch.tensor(0.0, device=mask_preds.device)

    if x_cat.shape[1] > 0:
        for idx in range(x_cat.shape[1]):
            categorical_feature_loss += categorical_loss_fn(cat_feature_preds[idx], x_cat[:, idx].long())
    if x_cont.shape[1] > 0:
        continuous_feature_loss += continuous_loss_fn(cont_feature_preds, x_cont)

    return mask_loss, categorical_feature_loss, continuous_feature_loss


def second_phase_step(
    model: nn.Module, batch: Tuple[torch.Tensor, torch.Tensor]
) -> torch.Tensor:
    """Forward step of VIME during the second phase

    Args:
        model (nn.Module): An instance of VIME
        batch (Tuple[torch.Tensor, torch.Tensor]): The input batch

    Returns:
        torch.Tensor: The predicted label (logit)
    """
    x, _ = batch
    
    return model(x).squeeze()


def second_phase_loss(
    y: torch.Tensor,
    y_hat: torch.Tensor,
    unlabeled_y_hat: torch.Tensor,
    consistency_loss_fn: nn.Module,
    task_loss_fn: nn.Module,
    K: int,
) -> Tuple[torch.Tensor, torch.Tensor]:
    """Calculate the second phase loss of VIME

    Args:
        y (torch.Tensor): The ground truth label
        y_hat (torch.Tensor): The predicted label
        unlabeled_y_hat (torch.Tensor): The predicted labels for the consistency regularization
        consistency_loss_fn (nn.Module): The loss function for the consistency regularization
        loss_fn (nn.Module): The loss function for the given task
        K (int): The number of perturbed samples for the consistency regularization
    Returns:
        Tuple[torch.Tensor, torch.Tensor, torch.Tensor]: The losses for the given task and consistency regularization and the ground truth labels
    """

    task_loss = task_loss_fn(y_hat, y)

    if len(unlabeled_y_hat) == 0:
        return task_loss, torch.tensor(0.0, device=y.device)

    consistency_len = K + 1
    
    # Select targets at intervals of consistency_len
    target = unlabeled_y_hat[::consistency_len]
    target = target.repeat_interleave(K, dim=0)

    # Select predictions that are not at intervals of consistency_len
    mask = torch.ones(len(unlabeled_y_hat), dtype=torch.bool, device=y.device)
    mask[::consistency_len] = False
    preds = unlabeled_y_hat[mask].view(-1, unlabeled_y_hat[mask].shape[-1]).squeeze()

    consistency_loss = consistency_loss_fn(preds, target)

    return task_loss, consistency_loss



import pandas as pd
from typing import Any, List, Union, Optional, Tuple
from numpy.typing import NDArray

import torch
from torch.utils.data import Dataset

import numpy as np
from ts3l.utils.vime_utils import VIMEConfig

class VIMEDataset(Dataset):
    def __init__(self,
                X: pd.DataFrame,
                Y: Optional[Union[NDArray[np.int_], NDArray[np.float64]]] = None, 
                config: Optional[VIMEConfig] = None, 
                unlabeled_data: Optional[pd.DataFrame] = None, 
                continuous_cols: Optional[List] = None, 
                category_cols: Optional[List] = None, 
                u_label: Any = -1, 
                is_second_phase: Optional[bool] = False,
                is_regression: Optional[bool] = False, 
        ) -> None:
        """A dataset class for VIME that handles labeled and unlabeled data.

        This class is designed to manage data for the VIME, accommodating both labeled and unlabeled datasets
        for self- and semi-supervised learning scenarios. It supports regression and classification tasks.

        Args:
            X (pd.DataFrame): DataFrame containing the features of the labeled data.
            Y (Union[NDArray[np.int_], NDArray[np.float64]], optional): Numpy array containing the labels for the data. 
                Use integers for classification labels and floats for regression targets. Defaults to None.
            config (Dict[str, Any]): The given hyperparameter set for VIME.
            unlabeled_data (pd.DataFrame): DataFrame containing the features of the unlabeled data, used for 
                self-supervised learning. Defaults to None.
            continuous_cols (List, optional): List of continuous columns. Defaults to None.
            category_cols (List, optional): List of categorical columns. Defaults to None.
            u_label (int, optional): The specifier for unlabeled sample. Defaults to -1.
            is_second_phase (bool): The flag that determines whether the dataset is for first phase or second phase learning. Default is False.
            is_regression (bool, optional): Flag indicating whether the task is regression (True) or classification (False).
                Defaults to False.
        """
        
        self.label = None
        
        if not is_second_phase:
            self.__getitem = self.__first_phase_get_item # type: ignore
            
        else:
            self.__getitem = self.__second_phase_get_item # type: ignore
            
            self.label_class = torch.FloatTensor if is_regression else torch.LongTensor
            
            if Y is not None:
            
                self.label = self.label_class(Y)
            
                if unlabeled_data is not None:
                    self.label = torch.concat((self.label, self.label_class([u_label for _ in range(len(unlabeled_data))])), dim=0) # type: ignore
                
                if not is_regression:
                    class_counts = [sum((self.label == i)) for i in set(self.label.numpy())]
                    num_samples = len(self.label)

                    class_weights = [num_samples/class_counts[i] for i in range(len(class_counts))]
                    self.weights = [class_weights[self.label[i]] for i in range(int(num_samples))]
                else:
                    self.weights = [1.0 for _ in range(len(X))]
                    if unlabeled_data is not None:
                        unlabeled_weight = len(X) / len(unlabeled_data)
                        self.weights.extend([unlabeled_weight for _ in range(len(unlabeled_data))])
        
        if unlabeled_data is not None:
            X = pd.concat([X, unlabeled_data])
            
        self.cont_data = torch.FloatTensor(X[continuous_cols].values)
        self.cat_data = torch.FloatTensor(X[category_cols].values)
        
        self.continuous_cols = continuous_cols
        self.category_cols = category_cols

        if config is not None:
            self.config = config
        
        self.u_label = u_label
    
    def __getitem__(self, idx: int) -> Union[Tuple[torch.Tensor, torch.Tensor], Tuple[torch.Tensor, torch.Tensor, torch.Tensor]]:
        """Retrieves the feature and label tensors for a given index. The structure of the label 
            varies depending on the learning phase: 

            - In the first phase, the label is a tuple containing a mask and the original feature tensor.
            - In the second phase, the label is either a token indicating unlabeled samples or the actual label for labeled samples.

        Args:
            idx (int): The index of the data point to retrieve.

        Returns:
            Union[Tuple[torch.Tensor, torch.Tensor], Tuple[torch.Tensor, torch.Tensor, torch.Tensor]]: The feature tensor and the label for the given index. 
            For first phase learning, label is a tuple of mask and original feature tensor.
            For second phase learning, label is a one of a token for unlabeled samples or a label of it.
        """
        return self.__getitem(idx)
    
    def __mask_generator(self, p_m, x):
        """Generate mask vector.
        
        Args:
            - p_m: corruption probability
            - x: feature matrix
            
        Returns:
            - mask: binary mask matrix 
        """
        mask = torch.bernoulli(torch.full(x.shape, p_m, device=x.device))

        return mask.unsqueeze(0)

    def __pretext_generator(self, m, x, empirical_dist) -> Tuple[torch.LongTensor, torch.FloatTensor]:  
        """Generate corrupted samples.
        
        Args:
            m: mask matrix
            x: feature matrix
            empirical_dist: empirical distribution matrix
            
        Returns:
            m_new: final mask matrix after corruption
            x_tilde: corrupted feature matrix
        """
        
        # Parameters
        dim = x.shape[0]
        
        # Randomly (and column-wise) shuffle data
        rand_idx = torch.randint(0, empirical_dist.size(0), (dim,), device=x.device)
        x_bar = empirical_dist[rand_idx, torch.arange(dim, device=x.device)]
        
        # Corrupt samples
        x_tilde = x * (1-m) + x_bar * m  
        # Define new mask matrix
        m_new = 1 * (x != x_tilde)
        
        if dim != 1:
            return m_new.squeeze(), x_tilde.squeeze()
        else:
            return m_new.squeeze().reshape(1), x_tilde.squeeze().reshape(1)
    
    def __generate_x_tildes(self, cat_samples: torch.Tensor, cont_samples:torch.Tensor) -> torch.Tensor:
        """Generate x_tilde for consistency regularization

        Args:
            cat_samples (torch.Tensor): The categorical features to generate x_tilde
            cont_samples (torch.Tensor): The continuous features to generate x_tilde

        Returns:
            torch.Tensor: x_tilde for consistency regularization
        """
        m_unlab = self.__mask_generator(self.config.p_m, cat_samples)
        cat_m_label, cat_x_tilde = self.__pretext_generator(m_unlab, cat_samples, self.cat_data)
        
        m_unlab = self.__mask_generator(self.config.p_m, cont_samples)
        cont_m_label, cont_x_tilde = self.__pretext_generator(m_unlab, cont_samples, self.cont_data)
        x_tilde = torch.concat((cat_x_tilde, cont_x_tilde)).float()
        
        return x_tilde
    
    def __first_phase_get_item(self, idx: int) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
        """Return a input and label pair

        Args:
            idx (int): The index of the data to sample

        Returns:
            Tuple[torch.Tensor, torch.Tensor, torch.Tensor]: A pair of input and label for first phase learning
        """
        cat_samples = self.cat_data[idx]
        m_unlab = self.__mask_generator(self.config.p_m, cat_samples)
        cat_m_label, cat_x_tilde = self.__pretext_generator(m_unlab, cat_samples, self.cat_data)

        cont_samples = self.cont_data[idx]
        m_unlab = self.__mask_generator(self.config.p_m, cont_samples)
        cont_m_label, cont_x_tilde = self.__pretext_generator(m_unlab, cont_samples, self.cont_data)

        m_label = torch.concat((cat_m_label, cont_m_label)).float()
        x_tilde = torch.concat((cat_x_tilde, cont_x_tilde)).float()

        x = torch.concat((cat_samples, cont_samples))

        return x_tilde, m_label, x

    def __second_phase_get_item(self, idx) -> Tuple[torch.Tensor, torch.Tensor]:
        """Return a input and label pair

        Args:
            idx (int): The index of the data to sample

        Returns:
            Tuple[torch.Tensor, torch.Tensor]: A pair of input and label for second phase learning
        """
        cat_samples = self.cat_data[idx]
        cont_samples = self.cont_data[idx]
        x = torch.concat((cat_samples, cont_samples)).squeeze()
        if self.label is not None:
            
            if self.label[idx] == self.u_label:
                _xs = [x]
                _xs.extend([self.__generate_x_tildes(cat_samples, cont_samples) for _ in range(self.config.K)])
                xs = torch.stack(_xs)
                return xs, self.label_class([self.u_label for _ in range(len(xs))])
            else:
                return x, self.label[idx].unsqueeze(0)
        else:
            return x, self.u_label
            
    def __len__(self):
        """Return the length of the dataset
        """
        return len(self.cat_data)

class VIMESecondPhaseCollateFN(object):
    """A callable class designed for batch processing, specifically tailored for the second phase learning with VIME. 
    It consolidates a batch of samples into a single dictionary with concatenated inputs and labels, suitable for model input.

    This class is meant to be used as a collate function in a DataLoader, where it efficiently organizes batch data 
    for training during second phase learning.
    """
    
    def __call__(self, batch):
        return torch.concat([x.unsqueeze(0) if x.ndim == 1 else x for x, _, in batch], dim=0), torch.concat([label for _, label in batch], dim=0)


import torch
from torch import nn

from typing import List, Tuple

class ReconstructionHead(nn.Module):
    def __init__(self, input_dim: int, num_continuous: int, cat_cardinality: List[int]):
        super().__init__()
        self.categorical_head = nn.ModuleList([nn.Linear(input_dim, card) for card in cat_cardinality])
        self.continuous_head = nn.Linear(input_dim, num_continuous)
        
    def forward(self, x: torch.Tensor)-> Tuple[List[torch.Tensor], torch.Tensor]:
        
        cat_preds = [linear(x) for linear in self.categorical_head]
        cont_preds = self.continuous_head(x)
        
        return cat_preds, cont_preds


from typing import Dict, Any, List
import pandas as pd
from sklearn import metrics
import torchmetrics
    
class RegressionMetric(object):
    def __init__(self, metric: str, metric_hparams: Dict[str, Any] = {}):
        """Regression Metric Wrapper for TabularS3L.

        Args:
            metric (str): A name of regression metric of sklearn.metrics or torchmetrics.functional.
            metric_hparams (_type_): A hyperparameters for the given metric. Default is an empty dictionary.
        """
        super(RegressionMetric, self).__init__()
        
        self.metric_hparams = metric_hparams
        
        if hasattr(metrics, metric):
            self.metric = getattr(metrics, metric)
            self.forward = self.sklearn_forward
        else:
            self.metric = getattr(torchmetrics.functional, metric)
            self.forward = self.torchmetrics_forward
            
        self.__name__ = self.metric.__name__
    
    def __call__(self, preds, target):
        return self.forward(preds, target)
    
    def sklearn_forward(self, preds, target):
        return self.metric(target, preds, **self.metric_hparams)

    def torchmetrics_forward(self, preds, target):
        return self.metric(preds, target, **self.metric_hparams)

class ClassificationMetric(object):
    def __init__(self, metric: str, metric_hparams: Dict[str, Any] = {}):
        """Classification Metric Wrapper for TabularS3L.

        Args:
            metric (str): A name of classification metric of sklearn.metrics or torchmetrics.functional.
            metric_hparams (_type_): A hyperparameters for the given metric. Default is an empty dictionary.
        """
        super(ClassificationMetric, self).__init__()
        
        self.metric_hparams = metric_hparams
        
        if hasattr(metrics, metric):
            self.metric = getattr(metrics, metric)
            self.forward = self.sklearn_forward
        else:
            self.metric = getattr(torchmetrics.functional, metric)
            self.forward = self.torchmetrics_forward
            
        self.__name__ = self.metric.__name__
    
    def __call__(self, preds, target):
        return self.forward(preds, target)
    
    def sklearn_forward(self, preds, target):
        return self.metric(target, preds.argmax(1), **self.metric_hparams)
    
    def torchmetrics_forward(self, preds, target):
        return self.metric(preds.argmax(1), target, **self.metric_hparams)
    
def get_category_cardinality(data: pd.DataFrame, category_cols: List[str]):
    """Calculate the number of unique values (cardinality) for each categorical column specified in a dataset.

    This function iterates over each column specified as categorical in 'category_cols', determines
    the unique elements in each of those columns, and appends the count of these unique elements
    to a list, which it returns.
    
    Args:
        data (pd.DataFrame): The dataset containing the categorical columns.
        category_cols (List[str]): A list of column names in 'data' that are considered categorical.

    Returns:
        List[int]: A list containing the counts of unique values for each categorical column specified.
                    Each element in the list corresponds to the number of unique values in the respective
                    column in 'category_cols'.
    """
    category_dims = []
    for col in category_cols:
        category_dims.append(len(set(data[col].values)))
    return category_dims


# Copyright (c) Minwook Kim
# Licensed under the MIT License.

""" Setup
"""
from setuptools import setup, find_packages
from codecs import open
from os import path
import pathlib

import pkg_resources

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

# requirements
with open(path.join(here, 'requirements.txt'), encoding='utf-8') as requirements_txt:
    install_requires = [
        str(requirement)
        for requirement
        in pkg_resources.parse_requirements(requirements_txt)
    ]

setup(
    name='ts3l',
    version='v0.60',
    description='A PyTorch Lightning-based library for self- and semi-supervised learning on tabular data.',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/Alcoholrithm/TabularS3L',
    author='Minwook Kim',
    author_email='kmiiiaa@pusan.ac.kr',

    # Note that this is a string of words separated by whitespace, not a list.
    keywords='tabular-data semi-supervised-learning self-supervised-learning VIME SubTab SCARF Denoising-AutoEncoder SwitchTab',
    packages=find_packages(),
    # The `include_package_data` parameter in the `setup()` function is used to specify whether to
    # include non-Python files (such as data files, configuration files, etc.) that are part of the
    # package when it is installed.
    include_package_data=False,
    install_requires=install_requires,
    python_requires='>=3.7',
)



from .datamodule import TS3LDataModule
from .misc import RegressionMetric, ClassificationMetric, get_category_cardinality
from .base_config import BaseConfig
from .embedding_utils import BaseEmbeddingConfig
from .backbone_utils import BaseBackboneConfig

__all__ = ["TS3LDataModule", "vime_utils", "subtab_utils", "scarf_utils", "RegressionMetric", "ClassificationMetric", "BaseConfig", "BaseEmbeddingConfig", "BaseBackboneConfig", "get_category_cardinality"]


import torch
import torch.nn as nn
from torch.nn import functional as F

import numpy as np
from numpy.typing import NDArray

from typing import Tuple, Union
from ts3l.models.common import TS3LModule
from ts3l.functional.subtab import arrange_tensors

from ts3l.utils import BaseEmbeddingConfig, BaseBackboneConfig
from ts3l.models.common import TS3LBackboneModule

class ShallowDecoder(nn.Module):
    def __init__(self,
                 hidden_dim : int,
                 out_dim : int
    ) -> None:
        super().__init__()

        self.net = nn.Linear(hidden_dim, out_dim)
    
    def forward(self, x : torch.Tensor) -> torch.Tensor:
        return self.net(x)

class Projector(nn.Module):
    def __init__(self, emb_dim, hidden_dim):
        super().__init__()
        self.projection_net = nn.Sequential(
            nn.Linear(emb_dim, hidden_dim),
            nn.LeakyReLU(),
            nn.Linear(hidden_dim, hidden_dim),
        )
    
    def forward(self, x):
        x = self.projection_net(x)
        return F.normalize(x, p = 2, dim = 1)
class SubTab(TS3LModule):

    def __init__(self,
                 embedding_config: BaseEmbeddingConfig,
                 backbone_config: BaseBackboneConfig,
                 output_dim: int,
                 projection_dim: int,
                 shuffle: bool,
                 n_subsets: int,
                 overlap_ratio: float,
                 mask_ratio: float,
                 noise_type: str,
                 noise_level: float,
                 **kwargs
    ) -> None:
        """Initialize SubTab

        Args:
            embedding_config (BaseEmbeddingConfig): Configuration for the embedding layer.
            backbone_config (BaseBackboneConfig): Configuration for the backbone network.
            output_dim (int): The dimensionality of the output.
            projection_dim (int): The dimensionality of the projector.
            shuffle (bool): Whether to shuffle the subsets. 
            n_subsets (int): The number of subsets to generate different views of the data.
            overlap_ratio (float): A hyperparameter that is to control the extent of overlapping between the subsets.
            mask_ratio (floats): Ratio of features to be masked as noise.
            noise_type (str): The type of noise to apply.
            noise_level (float): Intensity of Gaussian noise to be applied.
        """
        
        self.shuffle = shuffle
        
        self.n_subsets = n_subsets
        self.overlap_ratio = overlap_ratio
        
        self.mask_ratio = mask_ratio
        self.noise_type = noise_type
        self.noise_level = noise_level
        
        self.input_dim = embedding_config.input_dim
        
        self.emb_dim = embedding_config.emb_dim if hasattr(embedding_config, "emb_dim") else 1
        
        if backbone_config.name == "mlp" and embedding_config.name == "feature_tokenizer": # type: ignore
            self.column_idx = np.arange(self.input_dim * embedding_config.emb_dim) # type: ignore
            
        else:
            self.column_idx = np.arange(self.input_dim)
        
        
        self.n_feature_subset = self.input_dim // self.n_subsets
        self.subset_dim = self.n_feature_subset
        
        # Number of overlapping features between subsets
        self.n_overlap = int(self.overlap_ratio * self.n_feature_subset)
        self.n_overlap_dim = self.n_overlap
        
        if backbone_config.name == "mlp" and embedding_config.name == "feature_tokenizer": # type: ignore
            self.subset_dim = self.subset_dim * embedding_config.emb_dim # type: ignore
            self.n_overlap_dim = self.n_overlap_dim * embedding_config.emb_dim # type: ignore
        
        if embedding_config.name == "feature_tokenizer": # type: ignore
            self.__generate_subset_embedding = self.__post_embedding_subset # type: ignore
        else:
            self.__generate_subset_embedding = self.__pre_embedding_subset # type: ignore
        
        super(SubTab, self).__init__(embedding_config, backbone_config)
        
        self.projector = Projector(self.backbone_module.output_dim, projection_dim)
        self.decoder = ShallowDecoder(self.backbone_module.output_dim, self.embedding_module.input_dim)
        self.head = nn.Sequential(
            nn.Linear(self.backbone_module.output_dim, output_dim)
        )

    def _set_backbone_module(self, backbone_config):
        
        if hasattr(backbone_config, "input_dim"):
            backbone_config.input_dim = self.subset_dim + self.n_overlap_dim

        self.backbone_module = TS3LBackboneModule(backbone_config)
        
    @property
    def encoder(self) -> nn.Module:
        return self.backbone_module
    
    def __generate_subset_embedding(self,x: torch.Tensor):
        pass
    
    def __post_embedding_subset(self, x: torch.Tensor) -> torch.Tensor:
        x = self.embedding_module(x)
        x = self.__generate_subset(x)
        return x
    
    def __pre_embedding_subset(self, x: torch.Tensor) -> torch.Tensor:
        x = self.__generate_subset(x)
        x = self.embedding_module(x)
        return x
    
    def _first_phase_step(self, x : torch.Tensor) -> Tuple[torch.Tensor, torch.Tensor]:
        x = self.__generate_subset_embedding(x)
        x = self.encoder(x)
        projections = self.projector(x)
        x_recons = self.decoder(x)
        
        return projections, x_recons
    
    def _second_phase_step(self, 
                x : torch.Tensor,
                return_embeddings : bool = False) -> Union[Tuple[torch.Tensor, torch.Tensor], torch.Tensor]:
        x = self.__generate_subset_embedding(x)
        x = self.encoder(x)
        
        latent = arrange_tensors(x, self.n_subsets)
        latent = latent.reshape(x.shape[0] // self.n_subsets, self.n_subsets, -1).mean(1)
        out = self.head(latent)

        if return_embeddings:
            return out, latent
        return out
    
    def __generate_noisy_xbar(self, x : torch.Tensor) -> torch.Tensor:
        """Generates a noisy version of the input sample `x`.

        Args:
            x (torch.Tensor): The original sample.

        Returns:
            torch.Tensor: The noisy sample.
        """
        
        no = len(x)
        dim = self.n_feature_subset + self.n_overlap
        
        # Initialize corruption array
        x_bar = torch.zeros(x.shape).to(x.device)

        # Randomly (and column-wise) shuffle data
        if self.noise_type == "Swap":
            # Generate random permutations for all columns
            permutations = torch.stack([torch.randperm(no).reshape((-1, 1)) for _ in range(dim)], dim=1).to(x.device)
            permutations = permutations.repeat((1, 1, self.emb_dim))
            permutations = permutations.reshape(x.shape)

            # Use advanced indexing to permute the tensor
            x_bar = torch.gather(x, 0, permutations)
            
        elif self.noise_type == "Gaussian":
            noise = torch.normal(mean=0.0, std=self.noise_level, size=x.shape, device=x.device)
            x_bar = x + noise

        return x_bar
    
    def __generate_x_tilde(self, x: torch.Tensor, subset_column_idx: NDArray[np.int32]) -> torch.Tensor:
        """Generates a masked and potentially noisy subset of the input sample `x` based on the provided column indices.

        Args:
            x (torch.Tensor): The original sample.
            subset_column_idx (NDArray[np.int32]): Indices of columns to include in the subset.

        Returns:
            torch.Tensor: The processed subset of the sample.
        """

        x_bar = x[:, subset_column_idx]
        x_bar_noisy = self.__generate_noisy_xbar(x_bar)
        mask = torch.bernoulli(torch.full(x_bar.shape, self.mask_ratio, device=x_bar.device))
        x_bar = x_bar * (1 - mask) + x_bar_noisy * mask
        
        return x_bar
    
    def __generate_subset(self,
                        x : torch.Tensor,
    ) -> torch.Tensor:
        """Generates subsets of features from the input sample `x`, applying shuffling, noise, and masking as configured.

        Args:
            x (torch.Tensor): The batch of samples.

        Returns:
            torch.Tensor: A tensor containing the generated subsets for the batch.
        """

        permuted_order = np.random.permutation(self.n_subsets) if self.shuffle else np.arange(self.n_subsets)
        
        subset_column_indice_list = [self.column_idx[:(self.subset_dim + self.n_overlap_dim)]]
        subset_column_indice_list.extend([self.column_idx[range((i * self.subset_dim - self.n_overlap_dim), ((i + 1) * self.subset_dim))] for i in range(self.n_subsets)])
        
        subset_column_indice = np.array(subset_column_indice_list)
        subset_column_indice = subset_column_indice[permuted_order]
        
        if len(subset_column_indice) == 1:
            subset_column_indice = np.concatenate([subset_column_indice, subset_column_indice])

        x_tildes = torch.concat([self.__generate_x_tilde(x, subset_column_indice[i]) for i in range(self.n_subsets)]) # [subset1, subset2, ... ,  subsetN]

        return x_tildes


from dataclasses import dataclass, field
from ts3l.utils import BaseConfig

from typing import Any, List, Optional

@dataclass
class SwitchTabConfig(BaseConfig):
    """ Configuration class for initializing components of the SwitchTabLightning Module, including hyperparameters of SwitchTab,
    optimizers, learning rate schedulers, and loss functions, along with their respective hyperparameters.

    Inherits Attributes:
        task (str): Specify whether the problem is regression or classification.
        embedding_config (BaseEmbeddingConfig): Configuration for the embedding layer.
        backbone_config (BaseBackboneConfig): Configuration for the backbone network.
        output_dim (int): The dimension of output.
        loss_fn (str): Name of the loss function to be used. Must be an attribute of 'torch.nn'.
        loss_hparams (Dict[str, Any]): Hyperparameters for the loss function. Default is empty dictionary.
        metric (str): Name of the metric to be used. Must be an attribute of 'torchmetrics.functional' or 'sklearn.metrics'. Default is None.
        metric_hparams (Dict[str, Any]): Hyperparameters for the metric. Default is an empty dictionary.
        optim (str): Name of the optimizer to be used. Must be an attribute of 'torch.optim'. Default is 'AdamW'.
        optim_hparams (Dict[str, Any]): Hyperparameters for the optimizer. Default is {'lr': 0.0001, 'weight_decay': 0.00005}.
        scheduler (str): Name of the learning rate scheduler to be used. Must be an attribute of 'torch.optim.lr_scheduler' or None. Default is None.
        scheduler_hparams (Dict[str, Any]): Hyperparameters for the scheduler. Default is None, indicating no scheduler is used.
        initialization (str): The way to initialize neural network parameters. Default is 'kaiming_uniform'.
        random_seed (int): Seed for random number generators to ensure reproducibility. Defaults to 42.
        
    New Attributes:
        u_label (Any): The special token for unlabeled samples.
        corruption_rate (float): The proportion of features to be corrupted. Default is 0.3.
        alpha (float): A hyperparameter that is to control the trade-off between the reconstruction loss and task loss during first phase. Default is 1.0.
    Raises:
        ValueError: Inherited from `BaseConfig` to indicate that a configuration for the task, optimizer, scheduler, loss function, or metric is either invalid or not specified.
    """

    u_label: Any = field(default=-1)
    
    corruption_rate: float = field(default=0.3)
    
    alpha: float = field(default=1.0)

    def __post_init__(self):
        super().__post_init__()


from dataclasses import dataclass, field
from .base_backbone_config import BaseBackboneConfig
from typing import Union, List, Optional
from torch import nn

@dataclass
class MLPBackboneConfig(BaseBackboneConfig):
    input_dim: Optional[int] = field(default=None)
    hidden_dims: Union[int, List[int]] = field(default=128)
    output_dim: Optional[int] = field(default=None)
    n_hiddens: int = field(default=2)
    activation: str = field(default='ReLU')
    use_batch_norm: bool = field(default=True)

    def __post_init__(self):
        self.name = "mlp"
        
        if isinstance(self.hidden_dims, int):
            if self.n_hiddens > 1:
                self.hidden_dims = [self.hidden_dims for _ in range(self.n_hiddens - 1)]
            else:
                self.output_dim = self.hidden_dims
                self.hidden_dims = []
        
        if self.input_dim is None:
            raise TypeError("__init__ missing 1 required positional argument: 'input_dim'")
        
        if self.output_dim is None:
            self.output_dim = self.hidden_dims[-1]

        if not hasattr(nn, self.activation):
            raise ValueError(f"{self.activation} is not a valid activation of torch.nn")


import torch
from torch import nn
from ts3l.utils import BaseBackboneConfig

from ts3l.models.common.mlp import MLP
from ts3l.models.common.transformer_encoder import TransformerEncoder

class TS3LBackboneModule(nn.Module):
    def __init__(self, config: BaseBackboneConfig):
        super().__init__()
        self.config = config
        
        self.__set_backbone_network()
    
    def __set_backbone_network(self):
        if self.config.name == "mlp":
            self.backbone = MLP(**self.config.__dict__)
        elif self.config.name == "transformer":
            self.backbone = TransformerEncoder(**self.config.__dict__)
        
        self.output_dim = self.backbone.output_dim
    
    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return self.backbone(x)


import torch
from torch import nn

class TransformerEncoder(nn.Module):
    def __init__(self, 
                    d_model: int,
                    ffn_factor: float,
                    hidden_dim: int,
                    dropout_rate: float,
                    encoder_depth: int = 3,
                    n_head: int = 2,
                    **kwargs) -> None:
        """Initializes the encoder module used in the SwitchTab, employing an FT-Transformer architecture

        Args:
            cont_nums (int): The number of continuous features.
            category_dims (List[int]): A list of dimensions of the categorical features.
            ffn_factor (float): The scaling factor for the size of the feedforward network within the transformer blocks.
            hidden_dim (int): The dimensionality of the hidden layers within the network.
            dropout_rate (float): The dropout rate used within the encoder.
            encoder_depth (int, optional): The number of layers in the encoder. Defaults to 3.
            n_head (int, optional): The number of attention heads in the encoder. Defaults to 2.
        """
        super().__init__()
        
        if d_model % n_head != 0:
            divisors = [n for n in range(1, d_model + 1) if d_model % n == 0]
    
            # Find the closest divisor to the original num_heads
            closest_num_heads = min(divisors, key=lambda x: abs(x - n_head))
            
            if closest_num_heads != n_head:
                print(f"Adjusting num_heads from {n_head} to {closest_num_heads} (closest valid divisor of {d_model})")
            
            n_head = closest_num_heads
        
        self.encoder_layers = nn.ModuleList([
                nn.TransformerEncoderLayer(d_model=d_model, nhead=n_head, dim_feedforward=int(hidden_dim*ffn_factor), dropout=dropout_rate, batch_first=True)
                for _ in range(encoder_depth)
            ])

        self.output_dim = d_model
        
    def forward(self, x: torch.Tensor) -> torch.Tensor:

        for layer in self.encoder_layers:
            x = layer(x)
        cls_token = x[:, 0]
        
        return cls_token


from .mlp import MLP
from .misc import initialize_weights
from .embedding import TS3LEmbeddingModule
from .base_model import TS3LModule
from .backbone import TS3LBackboneModule

__all__ = ["MLP", "initialize_weights", "TS3LModule", "TS3LEmbeddingModule", "TS3LBackboneModule"]